#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import argparse
import sys

import paleomix.nodes.validation as validation


class ErrHandler:
    def __init__(self, quiet=False):
        self.quiet = quiet
        self.duplicate_reads = 0

    def __call__(self, chrom, pos, records, name, seq, qual):
        self.duplicate_reads += 1
        if self.quiet:
            return

        print("%s:%i -- %s %s %s:" % (chrom, pos, name, seq, qual))
        for filename, records in sorted(records.items()):
            print("    - %s:" % (filename,))

            for idx, record in enumerate(records, start=1):
                print("% 8i. " % (idx,), end="")

                if record.is_paired:
                    if record.is_read1:
                        print("Mate 1 read", end="")
                    elif record.is_read2:
                        print("Mate 2 read", end="")
                    else:
                        print("Unpaired read", end="")
                else:
                    print("Unpaired read", end="")

                try:
                    print(" with read-group %r" % (record.get_tag("RG"),), end="")
                except KeyError:
                    pass

                print()


def parse_args(argv):
    parser = argparse.ArgumentParser(prog="paleomix dupcheck")
    parser.add_argument("files", nargs="+", help="One or more input BAM files.")
    parser.add_argument(
        "--quiet",
        default=False,
        action="store_true",
        help="Only print the number of BAM records where 1 or "
        "more potential duplicates duplicates were "
        "identified.",
    )

    return parser.parse_args(argv)


def main(argv):
    """Main function; takes a list of arguments but excluding sys.argv[0]."""
    args = parse_args(argv)
    handler = ErrHandler(quiet=args.quiet)
    validation.check_bam_files(args.files, handler)

    if args.quiet:
        print("%i" % (handler.duplicate_reads,))
    else:
        print("Found %i record(s) with duplicates." % (handler.duplicate_reads,))

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import CommandNode
from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    apply_options,
)

import paleomix.common.fileutils as fileutils
import paleomix.common.versions as versions


_VERSION_CHECK = versions.Requirement(
    call=("AdapterRemoval", "--version"),
    search=r"ver. (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(2, 2, 0),
)


class SE_AdapterRemovalNode(CommandNode):
    def __init__(
        self, input_file, output_prefix, threads=1, options={}, dependencies=()
    ):
        # See below for parameters in common between SE/PE
        cmd = _get_common_parameters(threads=threads, options=options)

        # Prefix for output files, ensure that all end up in temp folder
        cmd.set_option("--basename", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s.gz"
        cmd.set_kwargs(
            TEMP_OUT_BASENAME=os.path.basename(output_prefix),
            OUT_SETTINGS=output_prefix + ".settings",
            OUT_MATE_1=output_tmpl % ("truncated",),
            OUT_DISCARDED=output_tmpl % ("discarded",),
        )

        cmd.set_option("--file1", "%(IN_READS_1)s")
        cmd.set_kwargs(IN_READS_1=input_file)

        apply_options(cmd, options)

        CommandNode.__init__(
            self,
            command=cmd.finalize(),
            threads=threads,
            description="trimming SE adapters from %s"
            % fileutils.describe_files(input_file),
            dependencies=dependencies,
        )


class PE_AdapterRemovalNode(CommandNode):
    def __init__(
        self,
        input_file_1,
        input_file_2,
        output_prefix,
        collapse=True,
        threads=1,
        options={},
        dependencies=(),
    ):
        cmd = _get_common_parameters(threads=threads, options=options)

        # Prefix for output files, to ensure that all end up in temp folder
        cmd.set_option("--basename", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s.gz"
        cmd.set_kwargs(
            TEMP_OUT_BASENAME=os.path.basename(output_prefix),
            OUT_SETTINGS=output_prefix + ".settings",
            OUT_READS_1=output_tmpl % ("pair1.truncated",),
            OUT_READS_2=output_tmpl % ("pair2.truncated",),
            OUT_SINGLETON=output_tmpl % ("singleton.truncated",),
            OUT_DISCARDED=output_tmpl % ("discarded",),
        )

        if collapse:
            cmd.set_option("--collapse")

            cmd.set_kwargs(
                OUT_COLLAPSED=output_tmpl % ("collapsed",),
                OUT_COLLAPSED_TRUNC=output_tmpl % ("collapsed.truncated",),
            )

        cmd.set_option("--file1", "%(IN_READS_1)s")
        cmd.set_option("--file2", "%(IN_READS_2)s")
        cmd.set_kwargs(IN_READS_1=input_file_1, IN_READS_2=input_file_2)

        apply_options(cmd, options)

        CommandNode.__init__(
            self,
            command=cmd.finalize(),
            threads=threads,
            description="trimming PE adapters from %s"
            % fileutils.describe_paired_files(input_file_1, input_file_2),
            dependencies=dependencies,
        )


def _get_common_parameters(options, threads=1):
    cmd = AtomicCmdBuilder("AdapterRemoval", CHECK_VERSION=_VERSION_CHECK)

    # Gzip compress FASTQ files
    cmd.set_option("--gzip")

    # Trim Ns at read ends
    cmd.set_option("--trimns", fixed=False)
    # Trim low quality scores
    cmd.set_option("--trimqualities", fixed=False)

    # Fix number of threads to ensure consistency when scheduling node
    cmd.set_option("--threads", threads)

    # Ensure that any user-specified list of adapters is tracked
    adapter_list = options.pop("--adapter-list", None)
    if adapter_list is not None:
        cmd.cmd.set_option("--adapter-list", "%(IN_ADAPTER_LIST)s")
        cmd.command.set_kwargs(IN_ADAPTER_LIST=adapter_list)

    for key in ("--trim5p", "--trim3p"):
        values = options.pop(key, None)
        if values is not None:
            cmd.add_value(key)
            if isinstance(values, list):
                for value in values:
                    cmd.add_value(value)
            else:
                cmd.add_value(values)

    return cmd



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from paleomix.common.utilities import grouper
from paleomix.common.formats.msa import MSA


_NUM_BLOCKS = 6
_BLOCK_SIZE = 10
_BLOCK_SPACING = 2
_MAX_NAME_LENGTH = 30
_NAME_ENDS_AT = 36
_LINE_SIZE = _NUM_BLOCKS * _BLOCK_SIZE + (_NUM_BLOCKS - 1) * _BLOCK_SPACING


def interleaved_phy(msa, add_flag=False, max_name_length=_MAX_NAME_LENGTH):
    MSA.validate(msa)
    header = "%i %i" % (len(msa), msa.seqlen())
    if add_flag:
        header += " I"
    result = [header, ""]

    padded_len = min(max_name_length, max(len(name) for name in msa.names())) + 2
    padded_len -= padded_len % -(_BLOCK_SIZE + _BLOCK_SPACING) + _BLOCK_SPACING

    streams = []
    spacing = " " * _BLOCK_SPACING
    for record in sorted(msa):
        name = record.name[:max_name_length]
        padding = (padded_len - len(name)) * " "

        lines = []
        line = [name, padding]
        for block in grouper(_BLOCK_SIZE, record.sequence, fillvalue=""):
            block = "".join(block)
            if sum(len(segment) for segment in line) >= _LINE_SIZE:
                lines.append("".join(line))
                line = [block]
            else:
                line.extend((spacing, block))

        lines.append("".join(line))
        streams.append(lines)

    for rows in zip(*streams):
        result.extend(row for row in rows)
        result.append("")
    result.pop()

    return "\n".join(result)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import CommandNode
from paleomix.atomiccmd.builder import AtomicCmdBuilder
from paleomix.atomiccmd.command import AtomicCmd

from paleomix.common.fileutils import reroot_path
import paleomix.common.versions as versions


_VERSION_REGEX = r"Version: (\d+)\.(\d+)(?:\.(\d+))?"

SAMTOOLS_VERSION = versions.Requirement(
    call=("samtools",), search=_VERSION_REGEX, checks=versions.GE(1, 3, 1)
)

BCFTOOLS_VERSION = versions.Requirement(
    call=("bcftools",), search=_VERSION_REGEX, checks=versions.GE(1, 4, 0)
)

TABIX_VERSION = versions.Requirement(
    call=("tabix",), search=_VERSION_REGEX, checks=versions.GE(1, 3, 1)
)


class TabixIndexNode(CommandNode):
    """Tabix indexes a BGZip compressed VCF or pileup file.

    The class currently supports the following presets:
        - vcf -- BGZipped VCF file.
        - pileup -- BGZipped pileup (non-binary) as produced by 'mpileup'.
    """

    def __init__(self, infile, preset="vcf", dependencies=()):
        assert infile.lower().endswith(".bgz")
        if preset == "pileup":
            call = ["tabix", "-s", 1, "-b", 2, "-e", 2]
        elif preset == "vcf":
            call = ["tabix", "-p", preset]
        else:
            assert False, "Unxpected preset: %r" % preset

        self._infile = infile
        cmd_tabix = AtomicCmd(
            call + ["%(TEMP_IN_VCFFILE)s"],
            TEMP_IN_VCFFILE=os.path.basename(infile),
            IN_VCFFILE=infile,
            OUT_TBI=infile + ".tbi",
            CHECK_TABIX=TABIX_VERSION,
        )

        CommandNode.__init__(
            self,
            description="creating tabix %s index for %s" % (preset, infile),
            command=cmd_tabix,
            dependencies=dependencies,
        )

    def _setup(self, config, temp):
        """See CommandNode._setup."""
        infile = os.path.abspath(self._infile)
        outfile = reroot_path(temp, self._infile)
        os.symlink(infile, outfile)

        CommandNode._setup(self, config, temp)

    def _teardown(self, config, temp):
        """See CommandNode._teardown."""
        os.remove(reroot_path(temp, self._infile))

        CommandNode._teardown(self, config, temp)


class FastaIndexNode(CommandNode):
    """Indexed a FASTA file using 'samtools faidx'."""

    def __init__(self, infile, dependencies=()):
        self._infile = infile
        cmd_faidx = AtomicCmd(
            ["samtools", "faidx", "%(TEMP_IN_FASTA)s"],
            TEMP_IN_FASTA=os.path.basename(infile),
            IN_FASTA=infile,
            OUT_TBI=infile + ".fai",
            CHECK_SAM=SAMTOOLS_VERSION,
        )

        CommandNode.__init__(
            self,
            description="creating FAI index for %s" % (infile,),
            command=cmd_faidx,
            dependencies=dependencies,
        )

    def _setup(self, config, temp):
        """See CommandNode._setup."""
        infile = os.path.abspath(self._infile)
        outfile = reroot_path(temp, self._infile)
        os.symlink(infile, outfile)

        CommandNode._setup(self, config, temp)

    def _teardown(self, config, temp):
        """See CommandNode._teardown."""
        os.remove(reroot_path(temp, self._infile))

        CommandNode._teardown(self, config, temp)


class BAMIndexNode(CommandNode):
    """Indexed a BAM file using 'samtools index'."""

    def __init__(self, infile, index_format=".bai", dependencies=()):
        if index_format == ".bai":
            samtools_call = ["samtools", "index", "%(IN_BAM)s", "%(OUT_IDX)s"]
        elif index_format == ".csi":
            samtools_call = ["samtools", "index", "-c", "%(IN_BAM)s", "%(OUT_IDX)s"]
        else:
            raise ValueError(
                "Unknown format type %r; expected .bai or .csi" % (index_format,)
            )

        command = AtomicCmd(
            samtools_call,
            IN_BAM=infile,
            OUT_IDX=infile + index_format,
            CHECK_SAM=SAMTOOLS_VERSION,
        )

        CommandNode.__init__(
            self,
            description="creating %s index for %s" % (index_format[1:].upper(), infile),
            command=command,
            dependencies=dependencies,
        )


def merge_bam_files_command(input_files):
    merge = AtomicCmdBuilder(
        ["samtools", "merge", "-u", "-"],
        OUT_STDOUT=AtomicCmd.PIPE,
        CHECK_VERSION=SAMTOOLS_VERSION,
    )

    merge.add_multiple_values(input_files)

    return merge.finalize()



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import subprocess

import pytest

import paleomix
import paleomix.main as main
import paleomix.tools.factory as factory


###############################################################################
###############################################################################


class ProcError(RuntimeError):
    pass


def check_run(call, *args, **kwargs):
    devnull = os.open(os.devnull, os.O_RDONLY)
    kwargs.setdefault("stdin", devnull)
    kwargs.setdefault("close_fds", True)
    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.PIPE

    returncode = kwargs.pop("expected_returncode", 0)

    proc = subprocess.Popen(call, *args, **kwargs)
    os.close(devnull)

    stdout, stderr = proc.communicate()
    if proc.returncode != returncode:
        raise ProcError(
            "Command returned %i: %r:\nSTDOUT: %r\nSTDERR: %r"
            % (proc.returncode, call, stdout, stderr)
        )

    return stdout.decode("utf-8", "replace"), stderr.decode("utf-8", "replace")


# Simple test of the paleomxi command
@pytest.mark.slow
def test_paleomix_command():
    stdout, stderr = check_run(["paleomix"])

    assert "PALEOMIX - pipelines and tools for NGS data analyses" in stdout
    assert stderr == ""


# Simple test that all commands can be executed
@pytest.mark.slow
@pytest.mark.parametrize("command", main._COMMANDS)
def test_factory__command_usage(command):
    cmd = factory.new(command)
    call = cmd.finalized_call

    stdout, stderr = check_run(call + ["--help"])

    assert stdout.startswith("usage: paleomix {}".format(command))
    assert not stderr


# Simple test that all commands support '--version'
@pytest.mark.slow
@pytest.mark.parametrize("arg", ("-v", "--version"))
@pytest.mark.parametrize("command", main._COMMANDS)
def test_factory__command_versions(arg, command):
    cmd = factory.new(command)
    call = cmd.finalized_call

    stdout, stderr = check_run(call + [arg])

    assert stdout.startswith("paleomix {}".format(command))
    assert stdout.endswith(" v{}\n".format(paleomix.__version__))
    assert not stderr



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import sys
import time

from paleomix.common.utilities import fragment, cumsum


_DESC = "Processed {Records} records ({Progress}) in {Time}, est. {Remaining} left. Last {RecordsDelta} records in {TimeDelta}, now at {Contig}: {Position}"
_FINAL = (
    "Processed {Records} records in {Time}. Last {RecordsDelta} records in {TimeDelta}"
)


class BAMTimer:
    def __init__(self, bamfile, desc=None, step=1e6, out=sys.stderr):
        self._bam = None
        self._out = out
        self._desc = desc
        self._step = int(step)
        self._count = 0
        self._last_count = 0
        self._last_time = time.time()
        self._start_time = self._last_time
        self._last_fract = -1.0

        self._bam_references = None

        self._total = 0.0
        self._counts = []
        if bamfile and bamfile.header.get("HD", {}).get("SO", "NA") == "coordinate":
            self._bam = bamfile
            self._bam_references = self._bam.references

            lengths = bamfile.lengths
            self._total = float(sum(lengths)) or 1.0
            self._counts.append(0)
            self._counts.extend(cumsum(lengths))

    def increment(self, count=1, read=None):
        self._count += count
        if (self._count - self._last_count) >= self._step:
            current_time = time.time()
            self._print(current_time, read)
            self._last_time = current_time
            self._last_count = self._count
        return self

    def finalize(self):
        self._print(time.time(), None)

    def _print(self, current_time, read):
        desc = _FINAL
        contig, position, progress, remaining = "NA", "NA", "NA", "NA"
        if read and not read.is_unmapped and self._bam:
            fraction = (read.pos + self._counts[read.tid]) / self._total
            if fraction >= self._last_fract:
                self._last_fract = fraction
                contig = self._bam_references[read.tid]
                position = self._format_int(read.pos + 1)
                progress = "%.2f%%" % (fraction * 100,)

                current_running = current_time - self._start_time
                remaining = self._format_time(
                    current_running / fraction - current_running
                )
                desc = _DESC
            else:
                print(
                    "File appears to be unsorted, cannot estimate progress",
                    file=self._out,
                )
                self._bam = None

        if self._desc:
            print("%s: " % self._desc, end="", file=self._out)

        print(
            desc.format(
                Records=self._format_int(self._count),
                RecordsDelta=self._format_int(self._count - self._last_count),
                Time=self._format_time(current_time - self._start_time),
                TimeDelta=self._format_time(current_time - self._last_time),
                Contig=contig,
                Position=position,
                Progress=progress,
                Remaining=remaining,
            ),
            file=self._out,
        )

    def _format_time(self, ftime):
        utc = time.gmtime(ftime)
        return "%02i:%02i:%02is" % (utc.tm_hour, utc.tm_min, utc.tm_sec)

    def _format_int(self, value):
        return (",".join(fragment(3, str(value)[::-1])))[::-1]



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import operator
import pickle
import sys

import pytest

import paleomix.common.versions as versions

###############################################################################
###############################################################################
# Check class


def test_check__func_must_be_callable():
    with pytest.raises(TypeError):
        versions.Check("FooBar", 3, 7, 5)


def test_check_str():
    obj = versions.Check("FooBar", operator.lt, 3, 7, 5)
    assert str(obj) == "FooBar"


###############################################################################
###############################################################################
# Check class -- hash and comparisons


def test_check__eq_same_func_desc_and_version():
    obj_1 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    obj_2 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    assert hash(obj_1) == hash(obj_2)
    assert obj_1 == obj_2


def test_check__not_eq_for_diff_func_same_desc_and_version():
    obj_1 = versions.Check("Desc {}", operator.gt, 1, 2, 3)
    obj_2 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    assert hash(obj_1) != hash(obj_2)
    assert obj_1 != obj_2


def test_check__not_eq_for_diff_desc_same_func_and_version():
    obj_1 = versions.Check("Desc1 {}", operator.lt, 1, 2, 3)
    obj_2 = versions.Check("Desc2 {}", operator.lt, 1, 2, 3)
    assert hash(obj_1) != hash(obj_2)
    assert obj_1 != obj_2


def test_check__not_eq_for_same_func_desc_diff_version():
    obj_1 = versions.Check("Desc {}", operator.lt, 1, 2, 3)
    obj_2 = versions.Check("Desc {}", operator.lt, 1, 3, 3)
    assert hash(obj_1) != hash(obj_2)
    assert obj_1 != obj_2


###############################################################################
###############################################################################
# EQ class


def test_eq__str__one_value():
    obj = versions.EQ(1)
    assert str(obj) == "v1"


def test_eq__str__two_values():
    obj = versions.EQ(2, 1)
    assert str(obj) == "v2.1"


def test_eq__check_values__equal():
    obj = versions.EQ(2, 3)
    assert obj((2, 3))


def test_eq__check_values__not_equal():
    obj = versions.EQ(2, 3)
    assert not obj((1, 3))
    assert not obj((2, 2))
    assert not obj((1, 4))


def test_eq__check_values__equal_truncated():
    obj = versions.EQ(2, 3)
    assert obj((2, 3, 1))


def test_eq__check_values__equal_too_few_values():
    obj = versions.EQ(2, 3)
    with pytest.raises(ValueError):
        obj((2,))


def test_eq__check_values__not_equal_too_few_values():
    obj = versions.EQ(2, 3)
    with pytest.raises(ValueError):
        obj((1,))


###############################################################################
###############################################################################
# GE class


def test_ge__str__one_value():
    obj = versions.GE(1)
    assert str(obj) == "at least v1"


def test_ge__str__two_values():
    obj = versions.GE(2, 1)
    assert str(obj) == "at least v2.1"


def test_ge__check_values__greater_than_or_equal():
    obj = versions.GE(2, 3)
    assert obj((2, 3))
    assert obj((2, 4))
    assert obj((3, 0))


def test_ge__check_values__not_greater_than_or_equal():
    obj = versions.GE(2, 3)
    assert not obj((1, 3))
    assert not obj((2, 2))


def test_ge__check_values__greater_than_or_equal_truncated():
    obj = versions.GE(2, 3)
    assert obj((2, 3, 1))
    assert obj((2, 4, 2))


def test_ge__check_values__equal_too_few_values():
    obj = versions.GE(2, 3)
    with pytest.raises(ValueError):
        obj((2,))


def test_ge__check_values__not_equal_too_few_values():
    obj = versions.GE(2, 3)
    with pytest.raises(ValueError):
        obj((1,))


###############################################################################
###############################################################################
# LT class


def test_lt__str__one_value():
    obj = versions.LT(1)
    assert str(obj) == "prior to v1"


def test_lt__str__two_values():
    obj = versions.LT(2, 1)
    assert str(obj) == "prior to v2.1"


def test_lt__check_values__less_than():
    obj = versions.LT(2, 3)
    assert obj((2, 2))
    assert obj((1, 9))


def test_lt__check_values__not_less_than():
    obj = versions.LT(2, 3)
    assert not obj((2, 3))
    assert not obj((2, 4))


def test_lt__check_values__less_than_truncated():
    obj = versions.LT(2, 3)
    assert obj((2, 2, 1))
    assert obj((2, 1, 2))


def test_lt__check_values__less_than_too_few_values():
    obj = versions.LT(2, 3)
    with pytest.raises(ValueError):
        obj((1,))


def test_lt__check_values__not_less_than_too_few_values():
    obj = versions.LT(2, 3)
    with pytest.raises(ValueError):
        obj((3,))


###############################################################################
###############################################################################
# Any class


def test_any__str():
    obj = versions.Any()
    assert str(obj) == "any version"


def test_lt__check_values__always_true():
    obj = versions.Any()
    assert obj((1,))
    assert obj((2, 3))
    assert obj((4, 5, 6))
    assert obj((5, 6, 7, 8))


###############################################################################
###############################################################################
# And class


def test_and__init__non_check_value():
    with pytest.raises(ValueError):
        versions.And(versions.LT(2), None)


###############################################################################
###############################################################################
# And class -- str


def test_and__str__single_item():
    obj = versions.And(versions.GE(1))
    assert str(obj) == "at least v1"


def test_and__str__two_items():
    obj_ge = versions.GE(1, 2)
    obj_lt = versions.LT(3, 4)
    obj = versions.And(obj_ge, obj_lt)

    assert str(obj) == "at least v1.2 and prior to v3.4"


def test_and__str__two_items__first_is_operator():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.LT(3, 4)
    obj = versions.And(obj_1, obj_2)

    assert str(obj) == "(at least v1.2 and prior to v2.0) and prior to v3.4"


def test_and__str__two_items__second_is_operator():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.Or(versions.GE(2, 0), versions.LT(3, 4))
    obj = versions.And(obj_1, obj_2)

    assert str(obj) == "at least v1.2 and (at least v2.0 or prior to v3.4)"


###############################################################################
###############################################################################
# And class -- check_version


def test_and__check_version__both_true():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.LT(2, 0)
    obj = versions.And(obj_1, obj_2)
    assert obj((1, 3))


def test_and__check_version__first_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.And(obj_1, obj_2)
    assert not obj((1, 3))


def test_and__check_version__second_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.And(obj_1, obj_2)
    assert not obj((2, 3))


def test_and__check_version__neither_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.And(obj_1, obj_2)
    assert not obj((2, 2))


_TRUNCATED_VERSIONS = (
    (versions.GE(1, 2), versions.LT(2, 0)),
    (versions.GE(1, 2, 2), versions.LT(2, 0)),
    (versions.GE(1, 2), versions.LT(2, 0, 1)),
)


@pytest.mark.parametrize("obj_1, obj_2", _TRUNCATED_VERSIONS)
def test_and__check_version__truncated(obj_1, obj_2):
    obj = versions.And(obj_1, obj_2)
    assert obj((1, 3, 3))


_INSUFFICIENT_VALUES = (
    (versions.GE(1, 2, 2), versions.LT(2, 0)),
    (versions.GE(1, 2), versions.LT(2, 0, 1)),
    (versions.GE(1, 2, 2), versions.LT(2, 0, 1)),
)


@pytest.mark.parametrize("obj_1, obj_2", _INSUFFICIENT_VALUES)
def test_and__check_version__insufficient_number_of_values(obj_1, obj_2):
    obj = versions.And(obj_1, obj_2)
    with pytest.raises(ValueError):
        obj((1, 3))


###############################################################################
###############################################################################
# Or class


def test_or__init__non_check_value():
    with pytest.raises(ValueError):
        versions.Or(versions.LT(2), None)


###############################################################################
###############################################################################
# Or class -- str


def test_or__str__single_item():
    obj = versions.Or(versions.GE(1))
    assert str(obj) == "at least v1"


def test_or__str__two_items():
    obj_ge = versions.GE(1, 2)
    obj_lt = versions.LT(3, 4)
    obj = versions.Or(obj_ge, obj_lt)

    assert str(obj) == "at least v1.2 or prior to v3.4"


def test_or__str__two_items__first_is_operator():
    obj_1 = versions.Or(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.LT(3, 4)
    obj = versions.Or(obj_1, obj_2)

    assert str(obj) == "(at least v1.2 or prior to v2.0) or prior to v3.4"


def test_or__str__two_items__second_is_operator():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.And(versions.GE(2, 0), versions.LT(3, 4))
    obj = versions.Or(obj_1, obj_2)

    assert str(obj) == "at least v1.2 or (at least v2.0 and prior to v3.4)"


###############################################################################
###############################################################################
# Or class -- check_version


def test_or__check_version__both_true():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.LT(2, 0)
    obj = versions.Or(obj_1, obj_2)
    assert obj((1, 3))


def test_or__check_version__first_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.Or(obj_1, obj_2)
    assert obj((1, 3))


def test_or__check_version__second_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.Or(obj_1, obj_2)
    assert obj((2, 3))


def test_or__check_version__neither_true():
    obj_1 = versions.And(versions.GE(1, 2), versions.LT(2, 0))
    obj_2 = versions.And(versions.GE(2, 3), versions.LT(3, 0))
    obj = versions.Or(obj_1, obj_2)
    assert not obj((2, 2))


@pytest.mark.parametrize("obj_1, obj_2", _TRUNCATED_VERSIONS)
def test_or__check_version__truncated(obj_1, obj_2):
    obj = versions.Or(obj_1, obj_2)
    assert obj((1, 3, 3))


_INSUFFICIENT_VALUES_OR = (
    (versions.GE(1, 2, 2), versions.LT(2, 0)),
    (versions.GE(1, 2, 2), versions.LT(2, 0, 1)),
)


@pytest.mark.parametrize("obj_1, obj_2", _INSUFFICIENT_VALUES_OR)
def test_or__check_version__insufficient_number_of_values(obj_1, obj_2):
    obj = versions.Or(obj_1, obj_2)
    with pytest.raises(ValueError):
        obj((1, 3))


def test_or__check_version__insufficient_number_of_values__is_lazy():
    obj_1 = versions.GE(1, 2)
    obj_2 = versions.LT(2, 0, 1)
    obj = versions.Or(obj_1, obj_2)
    assert obj((1, 3))


###############################################################################
###############################################################################
# RequirementObj -- constructor


def test_requirementobj__init__defaults():
    obj = versions.RequirementObj(
        call=("echo", "foo"), search=r"(\d+)\.(\d+)", checks=versions.Any()
    )

    assert obj.name == "echo"
    assert obj.priority == 0


def test_requirementobj__init__non_defaults():
    obj = versions.RequirementObj(
        call=("bash", "foo"),
        search=r"(\d+)\.(\d+)",
        checks=versions.Any(),
        name="A name",
        priority=17,
    )

    assert obj.name == "A name"
    assert obj.priority == 17


###############################################################################
###############################################################################
# RequirementObj -- version


def _echo_version(version, dst="stdout", returncode=0):
    tmpl = "import sys; sys.%s.write(%r); sys.exit(%s);"
    return (sys.executable, "-c", tmpl % (dst, version, returncode))


_PIPES = ("stderr", "stdout")

_VERSION_CALL_RESULTS = (
    (r"v(\d+)", (3,)),
    (r"v(\d+)\.(\d+)", (3, 5)),
    (r"v(\d+)\.(\d+)\.(\d+)", (3, 5, 2)),
)


@pytest.mark.parametrize("pipe", _PIPES)
@pytest.mark.parametrize("regexp, equals", _VERSION_CALL_RESULTS)
def test_requirementobj__version__call(pipe, regexp, equals):
    call = _echo_version("v3.5.2\n", dst=pipe)
    obj = versions.RequirementObj(call=call, search=regexp, checks=versions.Any())
    assert obj.version == equals


def test_requirementobj__version__version_str_not_found():
    call = _echo_version("A typical error\n")
    obj = versions.RequirementObj(
        call=call, search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )

    with pytest.raises(versions.VersionRequirementError):
        getattr(obj, "version")


def test_requirementobj__version__command_not_found():
    obj = versions.RequirementObj(
        call=("xyzabcdefoo",), search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )

    try:
        obj.version
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        # Should include OSError message
        assert "No such file or directory" in str(error)


def test_requirementobj__version__command_not_executable():
    obj = versions.RequirementObj(
        call=("./README.rst",), search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )

    try:
        obj.version
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        # Should include OSError message
        assert "[Errno 13]" in str(error)


def test_requirementobj__version__return_code_is_ignored():
    obj = versions.RequirementObj(
        _echo_version("v1.2.3", returncode=1),
        search=r"v(\d+)\.(\d+)",
        checks=versions.Any(),
    )
    assert obj.version == (1, 2)


def test_requirementobj__version__func_call():
    def _return_version():
        return "This is v5.3!"

    obj = versions.RequirementObj(
        call=_return_version, search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )
    assert obj.version == (5, 3)


def test_requirementobj__version__func_call_with_arguments():
    def _return_version(arg1, arg2):
        assert (arg1, arg2) == (2, "foo")
        return "This is v5.3!"

    obj = versions.RequirementObj(
        call=(_return_version, 2, "foo"), search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )
    assert obj.version == (5, 3)


@pytest.mark.parametrize(
    "message", ("UnsupportedClassVersionError", "UnsupportedClassVersionError v1.2.3")
)
def test_requirementobj__version__outdated_jre__with_or_without_version_str(message):
    obj = versions.RequirementObj(
        call=lambda: message, search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )

    try:
        obj.version
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        assert "upgrade your version of Java" in str(error)


###############################################################################
###############################################################################
# RequirementObj -- executable


def test_requirementobj__executable__no_cli_args():
    obj = versions.RequirementObj(
        call=["samtools"], search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )

    assert obj.executable == "samtools"


def test_requirementobj__executable__with_cli_arguments():
    obj = versions.RequirementObj(
        call=["samtools", "--version"], search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )

    assert obj.executable == "samtools"


def test_requirementobj__executable__function():
    obj = versions.RequirementObj(
        call=lambda: "v1.1", search=r"v(\d+)\.(\d+)", checks=versions.Any()
    )

    assert obj.executable is None


###############################################################################
###############################################################################
# RequirementObj -- __call__


class CheckCounted(versions.Check):
    def __init__(self, return_value=True, expected=(1, 1)):
        self.count = 0
        self.return_value = return_value
        versions.Check.__init__(self, "counted {}", operator.eq, *expected)

    def _do_check_version(self, current, reference):
        assert current == reference
        self.count += 1
        return self.return_value


def test_requirementobj__call__result_is_cached():
    counter = CheckCounted()
    obj = versions.RequirementObj(
        call=lambda: "v1.1.3", search=r"(\d)\.(\d)", checks=counter
    )

    obj()
    obj()

    assert counter.count == 1


def test_requirementobj__call__result_is_cached_unless_forced():
    counter = CheckCounted()
    obj = versions.RequirementObj(
        call=lambda: "v1.1.3", search=r"(\d)\.(\d)", checks=counter
    )

    obj()
    obj(force=True)

    assert counter.count == 2


def test_requirementobj__call__check_fails__function():
    expected = (
        "Version requirements not met for test#1: "
        "Expected at least v1.1, but found v1.0"
    )

    obj = versions.RequirementObj(
        call=lambda: "v1.0.3",
        search=r"(\d)\.(\d)",
        checks=versions.GE(1, 1),
        name="test#1",
    )
    try:
        obj()
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        assert str(error) == expected


def test_requirementobj__call__check_fails():
    expected = (
        "Version requirements not met for test#1: "
        "Expected at least v1.1, but found v1.0"
    )

    obj = versions.RequirementObj(
        call=_echo_version("v1.0.2"),
        search=r"(\d)\.(\d)",
        checks=versions.GE(1, 1),
        name="test#1",
    )
    try:
        obj()
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        assert str(error) == expected


def test_requirementobj__call__check_fails__jre_outdated():
    expected = (
        "Version could not be determined for test#1:\n"
        "    Command  = %s -c import sys; "
        "sys.stdout.write('UnsupportedClassVersionError'); sys.exit(0);\n"
        "\n"
        "The version of the Java Runtime Environment on this\n"
        "system is too old; please check the the requirement\n"
        "for the program and upgrade your version of Java.\n"
        "\n"
        "See the documentation for more information." % (sys.executable,)
    )

    value = "UnsupportedClassVersionError"
    obj = versions.RequirementObj(
        call=_echo_version(value),
        search=r"(\d)\.(\d)",
        checks=versions.GE(1, 1),
        name="test#1",
    )
    try:
        obj()
        assert False  # pragma: no coverage
    except versions.VersionRequirementError as error:
        assert str(error) == expected


###############################################################################
###############################################################################
# Pickling of checks

_CAN_PICKLE_VALUES = (
    versions.EQ(1, 2, 3),
    versions.GE(1, 2, 3),
    versions.LT(1, 2, 3),
    versions.Any(),
    versions.And(versions.EQ(1, 2, 3)),
    versions.Or(versions.GE(1, 2, 3)),
)


@pytest.mark.parametrize("obj", _CAN_PICKLE_VALUES)
def test_check__can_pickle(obj):
    pickle.dumps(obj)


###############################################################################
###############################################################################
# Requirement


def test_requirement__obj_is_cached_for_same_values():
    obj1 = versions.Requirement("echo", "", versions.LT(1))
    obj2 = versions.Requirement("echo", "", versions.LT(1))
    assert obj1 is obj2


def test_requirement__new_obj_if_call_differ():
    obj1 = versions.Requirement("echo", "", versions.LT(1))
    obj2 = versions.Requirement("true", "", versions.LT(1))
    assert obj1 is not obj2


def test_requirement__new_obj_if_search_differ():
    obj1 = versions.Requirement("echo", r"(\d+)", versions.LT(1))
    obj2 = versions.Requirement("echo", "", versions.LT(1))
    assert obj1 is not obj2


def test_requirement__new_obj_if_checks_differ():
    obj1 = versions.Requirement("echo", "", versions.GE(1))
    obj2 = versions.Requirement("echo", "", versions.LT(1))
    assert obj1 is not obj2


def test_requirement__same_obj_if_name_differ():
    obj1 = versions.Requirement("echo", "", versions.GE(1))
    assert obj1.name == "echo"
    obj2 = versions.Requirement("echo", "", versions.GE(1), name="foo")
    assert obj2.name == "foo"
    assert obj1 is obj2

    obj3 = versions.Requirement("echo", "", versions.GE(1), name="bar")
    assert obj3.name == "bar"
    assert obj2 is obj3

    obj4 = versions.Requirement("echo", "", versions.GE(1))
    assert obj3.name == "bar"
    assert obj3 is obj4


def test_requirement_highest_priority_overrides():
    obj1 = versions.Requirement("echo", "", versions.LT(1), priority=0)
    assert obj1.priority == 0
    obj2 = versions.Requirement("echo", "", versions.LT(1), priority=5)
    assert obj1 is obj2
    assert obj2.priority == 5


def test_requirement_highest_priority_retained():
    obj1 = versions.Requirement("echo", "", versions.LT(1), priority=5)
    assert obj1.priority == 5
    obj2 = versions.Requirement("echo", "", versions.LT(1), priority=0)
    assert obj1 is obj2
    assert obj2.priority == 5



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import os
import logging

import pysam

import paleomix.common.argparse as argparse

from paleomix.common.bedtools import read_bed_file, sort_bed_by_bamfile
from paleomix.common.fileutils import swap_ext


class BAMStatsError(RuntimeError):
    pass


def collect_readgroups(args, handle):
    readgroups = {None: {"SM": "<NA>", "LB": "<NA>"}}
    if args.ignore_readgroups:
        return readgroups

    for readgroup in handle.header.get("RG", ()):
        key_id = readgroup["ID"]
        sample = readgroup["SM"]
        library = readgroup["LB"]

        readgroups[key_id] = {"SM": sample, "LB": library}
    return readgroups


def collect_references(args, handle):
    if args.regions:
        lengths = collections.defaultdict(int)
        for region in args.regions:
            lengths[region.name] += region.end - region.start

        lengths = dict(lengths)
    elif handle.nreferences <= args.max_contigs:
        lengths = dict(zip(handle.references, handle.lengths))
    else:
        lengths = {"<Genome>": sum(handle.lengths)}

    return lengths


def collect_bed_regions(filename):
    regions = []
    name_cache = {}
    for record in read_bed_file(filename):
        if len(record) < 4:
            record.name = "%s*" % (record.contig,)

        record.contig = name_cache.get(record.contig, record.contig)
        record.name = name_cache.get(record.name, record.name)

        regions.append(record.freeze())

    return regions


def parse_arguments(argv, ext):
    prog = "paleomix %s" % (ext.strip("."),)
    usage = "%s [options] sorted.bam [out%s]" % (prog, ext)
    parser = argparse.ArgumentParser(prog=prog, usage=usage)

    parser.add_argument(
        "infile",
        metavar="BAM",
        help="Filename of a sorted BAM file. If set to '-' "
        "the file is read from STDIN.",
    )
    parser.add_argument(
        "outfile",
        metavar="OUTPUT",
        nargs="?",
        help="Filename of output table; defaults to name of "
        "the input BAM with a '%s' extension. If "
        "set to '-' the table is printed to STDOUT." % (ext,),
    )
    parser.add_argument(
        "--target-name",
        default=None,
        metavar="NAME",
        help="Name used for 'Target' column; defaults to the "
        "filename of the BAM file.",
    )
    parser.add_argument(
        "--regions-file",
        default=None,
        dest="regions_fpath",
        help="BED file containing regions of interest; %s "
        "is calculated only for these grouping by the "
        "name used in the BED file, or the contig name "
        "if no name has been specified for a record." % (ext.strip("."),),
    )
    parser.add_argument(
        "--max-contigs",
        default=100,
        type=int,
        help="The maximum number of contigs allowed in a BAM "
        "file. If this number is exceeded, the entire "
        "set of contigs is aggregated into one pseudo-"
        "contig named '<Genome>'. This is done to "
        "limit table sizes",
    )
    parser.add_argument(
        "--ignore-readgroups",
        default=False,
        action="store_true",
        help="Ignore readgroup information in reads, and only "
        "provide aggregated statistics; this is required "
        "if readgroup information is missing or partial",
    )
    parser.add_argument(
        "--overwrite-output",
        default=False,
        action="store_true",
        help="Overwrite output file if it it exists; by "
        "default, the script will terminate if the file "
        "already exists.",
    )

    args = parser.parse_args(argv)
    if not args.outfile:
        args.outfile = swap_ext(args.infile, ext)

    if args.ignore_readgroups:
        args.get_readgroup_func = _get_readgroup_ignored
    else:
        args.get_readgroup_func = _get_readgroup

    if not args.target_name:
        if args.infile == "-":
            args.target_name = "<STDIN>"
        else:
            args.target_name = os.path.basename(args.infile)

    if os.path.exists(args.outfile) and not args.overwrite_output:
        parser.error(
            "Destination filename already exists (%r); use option "
            "--overwrite-output to allow overwriting of this file." % (args.outfile,)
        )

    return args


def main_wrapper(process_func, argv, ext):
    log = logging.getLogger(__name__)
    args = parse_arguments(argv, ext)
    args.regions = None
    if args.regions_fpath:
        try:
            args.regions = collect_bed_regions(args.regions_fpath)
        except ValueError as error:
            log.error("Failed to parse BED file %r: %s", args.regions_fpath, error)
            return 1

    log.info("Opening %r", args.infile)
    with pysam.AlignmentFile(args.infile) as handle:
        sort_order = handle.header.get("HD", {}).get("SO")
        if sort_order is None:
            log.warning("BAM file %r is not marked as sorted!", args.infile)
        elif sort_order != "coordinate":
            log.error(
                "BAM file %r is %s-sorted, but coordinate-sorting is required",
                args.infile,
                sort_order,
            )
            return 1

        sort_bed_by_bamfile(handle, args.regions)
        return process_func(handle, args)


def _get_readgroup(record):
    try:
        return record.get_tag("RG")
    except KeyError:
        return None


def _get_readgroup_ignored(_):
    return None



#!/usr/bin/python3
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os
import shutil
import string
import tarfile

import pysam

import paleomix
import paleomix.common.fileutils as fileutils
import paleomix.common.logging
import paleomix.pipelines.bam.mkfile as bam_mkfile
import paleomix.pipelines.zonkey.config as zonkey_config
import paleomix.pipelines.zonkey.database as database
import paleomix.pipelines.zonkey.parts.common as common_nodes
import paleomix.pipelines.zonkey.parts.mitochondria as mitochondria
import paleomix.pipelines.zonkey.parts.nuclear as nuclear
import paleomix.pipelines.zonkey.parts.report as report
import paleomix.pipelines.zonkey.parts.summary as summary
import paleomix.yaml
from paleomix.common.formats.fasta import FASTA
from paleomix.nodes.raxml import RAxMLRapidBSNode
from paleomix.nodes.samtools import BAMIndexNode
from paleomix.pipeline import Pypeline


def run_pipeline(config, nodes, msg):
    pipeline = Pypeline(config)
    pipeline.add_nodes(nodes)

    paleomix.common.logging.initialize(
        log_level=config.log_level, log_file=config.log_file, name="zonkey"
    )

    logger = logging.getLogger(__name__)
    logger.info(msg)

    if config.list_executables:
        pipeline.print_required_executables()
        return True
    elif config.list_output_files:
        pipeline.print_output_files()
        return True
    elif config.list_input_files:
        pipeline.print_input_files()
        return True

    return pipeline.run(max_threads=config.max_threads, dry_run=config.dry_run)


def build_plink_nodes(config, data, root, bamfile, dependencies=()):
    plink = {"root": os.path.join(root, "results", "plink")}

    ped_node = nuclear.BuildTPEDFilesNode(
        output_root=plink["root"],
        table=config.database.filename,
        downsample=config.downsample_to,
        bamfile=bamfile,
        dependencies=dependencies,
    )

    for postfix in ("incl_ts", "excl_ts"):
        plink[postfix] = nuclear.BuildBEDFilesNode(
            output_prefix=os.path.join(plink["root"], postfix),
            tfam=os.path.join(plink["root"], "common.tfam"),
            tped=os.path.join(plink["root"], postfix + ".tped"),
            plink_parameters=config.database.settings["Plink"],
            dependencies=(ped_node,),
        )

    return plink


def build_admixture_nodes(config, data, root, plink):
    nodes = []
    for postfix in ("incl_ts", "excl_ts"):
        bed_node = plink[postfix]

        admix_root = os.path.join(root, "results", "admixture")
        report_root = os.path.join(root, "figures", "admixture")
        for k_groups in sorted(data.groups):
            replicates = []

            input_file = os.path.join(plink["root"], postfix + ".bed")
            for replicate in range(config.admixture_replicates):
                output_root = os.path.join(admix_root, "%02i" % (replicate,))

                node = nuclear.AdmixtureNode(
                    input_file=input_file,
                    output_root=output_root,
                    k_groups=k_groups,
                    groups=data.groups[k_groups],
                    dependencies=(bed_node,),
                )

                replicates.append(node)

            node = nuclear.SelectBestAdmixtureNode(
                replicates=replicates, output_root=admix_root
            )

            if config.admixture_only:
                nodes.append(node)
            else:
                samples = os.path.join(root, "figures", "samples.txt")
                plot = nuclear.AdmixturePlotNode(
                    input_file=os.path.join(
                        admix_root, "%s.%i.Q" % (postfix, k_groups)
                    ),
                    output_prefix=os.path.join(
                        report_root, "%s_k%i" % (postfix, k_groups)
                    ),
                    samples=samples,
                    order=data.sample_order,
                    dependencies=node,
                )

                nodes.append(plot)

    return nodes


def build_treemix_nodes(config, data, root, plink):
    tmix_root = os.path.join(root, "results", "treemix")

    nodes = []
    for postfix in ("incl_ts", "excl_ts"):
        plink_prefix = os.path.join(plink["root"], postfix)
        plink_nodes = plink[postfix]

        freq_node = nuclear.BuildFreqFilesNode(
            output_prefix=plink_prefix,
            input_prefix=os.path.join(plink["root"], postfix),
            tfam=os.path.join(plink["root"], "common.tfam"),
            parameters=config.database.settings["Plink"],
            dependencies=plink_nodes,
        )

        tmix_prefix = os.path.join(tmix_root, postfix)
        tmix_file_node = nuclear.FreqToTreemixNode(
            input_file=plink_prefix + ".frq.strat.gz",
            output_file=tmix_prefix + ".gz",
            dependencies=(freq_node,),
        )

        k_snps = config.treemix_k
        if not k_snps:
            k_snps = (
                "n_sites_%s" % (postfix,),
                os.path.join(plink["root"], "common.summary"),
            )

        for n_migrations in (0, 1):
            n_prefix = "%s.%i" % (tmix_prefix, n_migrations)

            tmix_node = nuclear.TreemixNode(
                data=data,
                input_file=tmix_prefix + ".gz",
                output_prefix=n_prefix,
                m=n_migrations,
                k=k_snps,
                outgroup=config.treemix_outgroup,
                dependencies=(tmix_file_node,),
            )

            samples = os.path.join(root, "figures", "samples.txt")
            output_prefix = os.path.join(
                root, "figures", "treemix", "%s_%i" % (postfix, n_migrations)
            )
            plot_node = nuclear.PlotTreemixNode(
                samples=samples,
                prefix=n_prefix,
                output_prefix=output_prefix,
                dependencies=(tmix_node,),
            )

            nodes.append(plot_node)

    return nodes


def build_pca_nodes(config, data, root, plink):
    pca_root = os.path.join(root, "results", "pca")

    nodes = []
    for postfix in ("incl_ts", "excl_ts"):
        plink_prefix = os.path.join(plink["root"], postfix)
        plink_nodes = plink[postfix]

        pca_prefix = os.path.join(pca_root, postfix)
        pca_node = nuclear.SmartPCANode(
            input_prefix=plink_prefix,
            output_prefix=pca_prefix,
            nchroms=data.settings["NChroms"],
            dependencies=plink_nodes,
        )

        samples = os.path.join(root, "figures", "samples.txt")
        pca_plots = os.path.join(root, "figures", "pca", postfix)
        pca_plot_node = nuclear.PlotPCANode(
            samples=samples,
            prefix=pca_prefix,
            output_prefix=pca_plots,
            dependencies=pca_node,
        )

        nodes.append(pca_plot_node)

    return nodes


def build_coverage_nodes(contigs, mapping, root, nuc_bam, dependencies=()):
    output_prefix = os.path.join(root, "figures", "coverage", "coverage")

    return (
        nuclear.PlotCoverageNode(
            contigs=contigs,
            mapping=mapping,
            input_file=nuc_bam,
            output_prefix=output_prefix,
            dependencies=dependencies,
        ),
    )


def build_mito_nodes(config, root, bamfile, dependencies=()):
    if config.database.mitochondria is None:
        log = logging.getLogger(__name__)
        log.warning("No MT sequences in Zonkey database; cannot perform MT analysis")
        return ()

    samples = os.path.join(root, "figures", "samples.txt")

    mt_prefix = os.path.join(root, "results", "mitochondria", "sequences")
    alignment = mitochondria.MitoConsensusNode(
        database=config.database.filename,
        bamfile=bamfile,
        output_prefix=mt_prefix,
        dependencies=dependencies,
    )

    raxml_template = os.path.join(root, "results", "mitochondria", "raxml_%s")
    phylo = RAxMLRapidBSNode(
        input_alignment=mt_prefix + ".phy",
        output_template=raxml_template,
        model="GTRGAMMA",
        replicates=100,
        dependencies=(alignment,),
    )

    output_prefix = os.path.join(root, "figures", "mitochondria", "mito_phylo")
    trees = mitochondria.DrawPhylogenyNode(
        samples=samples,
        treefile=raxml_template % ("bestTree",),
        bootstraps=raxml_template % ("bootstrap",),
        output_prefix=output_prefix,
        dependencies=(phylo,),
    )

    return (trees,)


def build_pipeline(config, root, nuc_bam, mito_bam, cache):
    nodes = []
    sample_tbl = os.path.join(root, "figures", "samples.txt")
    samples = common_nodes.WriteSampleList(config=config, output_file=sample_tbl)

    if nuc_bam is not None:
        nuc_bam, nuc_bam_info = nuc_bam["Path"], nuc_bam["Info"]

        # When not sampling, BuildTPED relies on indexed access to ease
        # processing of one chromosome at a time. The index is further required
        # for idxstats used by the PlotCoverageNode.
        index = cache.get(nuc_bam)
        if index is None:
            index = cache[nuc_bam] = BAMIndexNode(infile=nuc_bam)

        plink = build_plink_nodes(
            config, config.database, root, nuc_bam, dependencies=(samples, index)
        )

        nodes.extend(build_admixture_nodes(config, config.database, root, plink))

        if not config.admixture_only:
            nodes.extend(
                build_coverage_nodes(
                    contigs=config.database.contigs,
                    mapping=nuc_bam_info.nuclear_contigs,
                    root=root,
                    nuc_bam=nuc_bam,
                    dependencies=(index,),
                )
            )
            nodes.extend(build_pca_nodes(config, config.database, root, plink))
            nodes.extend(build_treemix_nodes(config, config.database, root, plink))

    if mito_bam is not None and not config.admixture_only:
        index = cache.get(mito_bam)
        if index is None:
            index = cache[mito_bam] = BAMIndexNode(infile=mito_bam)

        nodes.extend(
            build_mito_nodes(config, root, mito_bam, dependencies=(samples, index))
        )

    if not config.admixture_only:
        nodes.append(
            report.ReportNode(config, root, nuc_bam, mito_bam, dependencies=nodes)
        )

    return nodes


def run_admix_pipeline(config):
    log = logging.getLogger(__name__)
    log.info("Building %i Zonkey pipeline(s):", len(config.samples))
    config.temp_root = os.path.join(config.destination, "temp")
    if not config.dry_run:
        fileutils.make_dirs(config.temp_root)

    cache = {}
    nodes = []
    items = iter(config.samples.items())
    for idx, (name, sample) in enumerate(sorted(items), start=1):
        root = sample["Root"]
        nuc_bam = sample["Files"].get("Nuc")
        mito_bam = sample["Files"].get("Mito")

        genomes = []
        if mito_bam:
            genomes.append("MT")
        if nuc_bam:
            genomes.append("Nuclear")

        log.info("  %i. %s: %s DNA", idx, name, " and ".join(genomes))

        nodes.extend(build_pipeline(config, root, nuc_bam, mito_bam, cache))

    if config.multisample and not config.admixture_only:
        nodes = [summary.SummaryNode(config, nodes)]

    if not run_pipeline(config, nodes, "Running Zonkey"):
        return 1


def setup_mito_mapping(config):
    genomes_root = os.path.join(config.destination, "genomes")
    if not os.path.exists(genomes_root):
        fileutils.make_dirs(genomes_root)

    mkfile_fpath = os.path.join(config.destination, "makefile.yaml")

    filenames = [mkfile_fpath]
    for name, record in sorted(config.database.mitochondria.items()):
        filenames.append(os.path.join(genomes_root, "%s.fasta" % (record.name,)))

    existing_filenames = [
        filename for filename in filenames if os.path.exists(filename)
    ]

    # A bit strict, but avoid accidential overwrites
    if existing_filenames:
        log = logging.getLogger(__name__)
        log.error("Output file(s) already exists, cannot proceed:")
        for filename in sorted(existing_filenames):
            log.error(" - %r", filename)

        return 1

    with open(mkfile_fpath, "w") as mkfile:
        mkfile.write(
            bam_mkfile.build_makefile(add_prefix_tmpl=False, add_sample_tmpl=False)
        )

        mkfile.write("\n\nPrefixes:\n")

        for name, record in sorted(config.database.mitochondria.items()):
            if "EXCLUDE" in record.meta.upper():
                continue

            mkfile.write("  %s:\n" % (record.name,))
            mkfile.write("    Path: genomes/%s.fasta\n" % (record.name,))

            info = config.database.samples.get(record.name)
            if info is not None:
                mkfile.write("    # Species: %s\n" % (info.get("Species", "NA"),))
                mkfile.write("    # Sex: %s\n" % (info.get("Sex", "NA"),))
                mkfile.write(
                    "    # Publication: %s\n" % (info.get("Publication", "NA"),)
                )
                mkfile.write("    # Sample ID: %s\n" % (info.get("SampleID", "NA"),))

            mkfile.write("\n")

            fasta_fpath = os.path.join(genomes_root, "%s.fasta" % (record.name,))

            with open(fasta_fpath, "w") as fasta_handle:
                record = FASTA(
                    name=record.name,
                    meta=None,
                    sequence=record.sequence.replace("-", ""),
                )

                record.write(fasta_handle)

        mkfile.write("\n")

    return 0


def setup_example(config):
    root = os.path.join(config.destination, "zonkey_pipeline")
    log = logging.getLogger(__name__)
    log.info("Copying example project to %r", root)

    with tarfile.TarFile(config.database.filename) as tar_handle:
        example_files = []
        existing_files = []
        for member in tar_handle.getmembers():
            if os.path.dirname(member.name) == "examples" and member.isfile():
                example_files.append(member)

                destination = fileutils.reroot_path(root, member.name)
                if os.path.exists(destination):
                    existing_files.append(destination)

        if existing_files:
            log.error("Output files already exist at destination:")
            for filename in sorted(existing_files):
                log.error(" - %r", filename)
            return 1
        elif not example_files:
            log.error(
                "Sample database %r does not contain example data; cannot proceed.",
                config.database.filename,
            )
            return 1

        if not os.path.exists(root):
            fileutils.make_dirs(root)

        for member in example_files:
            destination = fileutils.reroot_path(root, member.name)
            src_handle = tar_handle.extractfile(member)
            with open(destination, "wb") as out_handle:
                shutil.copyfileobj(src_handle, out_handle)

    log.info("Sucessfully saved example data in %r", root)

    return 0


def _process_samples(config):
    log = logging.getLogger(__name__)
    for name, info in sorted(config.samples.items()):
        files = {}

        if name == "-":
            log.info("Validating unnamed sample")
        else:
            log.info("Validating sample %r", name)

        for filename in info.pop("Files"):
            filetype = config.database.validate_bam(filename)
            if not filetype:
                log.error("File is not a valid BAM file: %r" % (filename,))
                return False

            if filetype.is_nuclear and filetype.is_mitochondrial:
                if "Nuc" in files:
                    log.error("Two nuclear BAMs specified!")
                    return False
                elif "Mito" in files:
                    log.warning(
                        "Nuclear + mitochondrial BAM and mitochondrial BAM specified; "
                        "the mitochondrial genome in the combined BAM will not be used!"
                    )

                files["Nuc"] = {"Path": filename, "Info": filetype}
                files.setdefault("Mito", filename)
            elif filetype.is_nuclear:
                if "Nuc" in files:
                    log.error("Two nuclear BAMs specified!")
                    return False

                files["Nuc"] = {"Path": filename, "Info": filetype}
            elif filetype.is_mitochondrial:
                if "Mito" in files:
                    log.error("Two nuclear BAMs specified!")
                    return False

                files["Mito"] = filename
            else:
                log.error(
                    "BAM does not contain usable nuclear or mitochondrial contigs: %r",
                    filename,
                )
                return False

        config.samples[name]["Files"] = files

    return True


def _read_sample_table(config, filename):
    """Parses a 2 - 3 column tab-seperated table containing, on each row, a
    name to be used for a sample in the first row, and then the paths two
    either one or to two BAM files, which must represent a single nuclear or
    a single mitochondrial alignment (2 columns), or both (3 columns).
    """
    log = logging.getLogger(__name__)
    log.info("Reading table of samples from %r", filename)
    valid_characters = frozenset(string.ascii_letters + string.digits + ".-_")

    samples = config.samples = {}
    with fileutils.open_ro(filename) as handle:
        for linenum, line in enumerate(handle, start=1):
            if not line.strip() or line.lstrip().startswith("#"):
                continue

            fields = [_f for _f in map(str.strip, line.split("\t")) if _f]
            if len(fields) not in (2, 3):
                log.error(
                    "Error reading sample table (%r) at line %i: Expected 2 or 3 "
                    "columns, found %i; please correct file before continuing.",
                    filename,
                    linenum,
                    len(fields),
                )
                return

            name = fields[0]
            invalid_letters = frozenset(name) - valid_characters
            if invalid_letters:
                log.error(
                    "Error reading sample table (%r) at line %i: Sample name contains "
                    "illegal character(s). Only letters, numbers, and '-', '_', and "
                    "'.' are allowed, but found %r in name %r ",
                    filename,
                    linenum,
                    "".join(invalid_letters),
                    name,
                )
                return
            elif name in samples:
                log.error(
                    "Duplicate name %r in sample table; names must be unique!", name
                )
                return

            samples[name] = {
                "Root": os.path.join(config.destination, name),
                "Files": fields[1:],
            }

    return True


def finalize_run_config(parser, args):
    log = logging.getLogger(__name__)
    if args.command in ("run", "dryrun") and not (1 <= len(args.files) <= 3):
        parser.print_usage()
        return

    args.dry_run |= args.command == "dryrun"
    args.multisample = False

    known_samples = set(args.database.samples) | set(("Sample",))
    unknown_samples = set(args.treemix_outgroup) - known_samples
    if unknown_samples:
        log.error(
            "Argument --treemix-outgroup includes unknown sample(s): %s; known "
            "samples are %s. Note that names are case-sensitive."
            ", ".join(map(repr, sorted(unknown_samples))),
            ", ".join(map(repr, sorted(known_samples))),
        )
        return

    if len(args.files) == 1:
        args.files.append(fileutils.swap_ext(args.files[0], ".zonkey"))

    if len(args.files) == 2:
        filename, args.destination = args.files

        if os.path.exists(args.destination) and not os.path.isdir(args.destination):
            log.error("Destination %r is not a directory", args.destination)
            return
        elif not os.path.isfile(filename):
            log.error("Not a valid filename: %r", filename)
            return
        elif _is_bamfile(filename):
            args.samples = {"-": {"Root": args.destination, "Files": [filename]}}
        else:
            args.multisample = True
            if not _read_sample_table(args, filename):
                return
    elif len(args.files) == 3:
        filename_1, filename_2, args.destination = args.files

        args.samples = {
            "-": {"Root": args.destination, "Files": [filename_1, filename_2]}
        }
    else:
        raise RuntimeError("Unexpected number of arguments: %r" % (args.files,))

    # Identify (mito or nuc?) and validate BAM files provided by user
    if not _process_samples(args):
        return

    return args


def _is_bamfile(filename):
    """Returns true if a file is a BAM file, false otherwise.
    """
    try:
        with pysam.AlignmentFile(filename, "rb"):
            return True
    except ValueError:
        return False
    except IOError:
        return False


def main(argv):
    parser, run_parser = zonkey_config.build_parser()
    if not argv:
        parser.print_help()
        return

    args = parser.parse_args(argv)
    log = logging.getLogger(__name__)

    try:
        args.database = database.ZonkeyDB(args.database)
    except database.ZonkeyDBError as error:
        log.error("Error reading database %r: %s", args.database, error)
        return

    if args.command in ("run", "dryrun"):
        args = finalize_run_config(run_parser, args)
        if args is not None:
            return run_admix_pipeline(args)
    elif args.command == "mito":
        return setup_mito_mapping(args)
    elif args.command == "example":
        return setup_example(args)

    return 1



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.node import CommandNode, NodeError
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    apply_options,
)
from paleomix.atomiccmd.sets import ParallelCmds
from paleomix.nodes.bwa import (
    _get_node_description,
    _new_cleanup_command,
    _get_max_threads,
)

import paleomix.common.versions as versions


BOWTIE2_VERSION = versions.Requirement(
    call=("bowtie2", "--version"),
    search=r"version (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(2, 3, 0),
)


class Bowtie2IndexNode(CommandNode):
    def __init__(self, input_file, prefix=None, dependencies=()):
        prefix = prefix if prefix else input_file
        builder = _bowtie2_template(
            ("bowtie2-build"),
            prefix,
            iotype="OUT",
            IN_FILE=input_file,
            TEMP_OUT_PREFIX=os.path.basename(prefix),
            CHECK_VERSION=BOWTIE2_VERSION,
        )

        builder.add_value("%(IN_FILE)s")
        # Destination prefix, in temp folder
        builder.add_value("%(TEMP_OUT_PREFIX)s")

        CommandNode.__init__(
            self,
            command=builder.finalize(),
            description="creating Bowtie2 index for %s" % (input_file,),
            dependencies=dependencies,
        )


class Bowtie2Node(CommandNode):
    def __init__(
        self,
        input_file_1,
        input_file_2,
        output_file,
        reference,
        prefix,
        threads=2,
        log_file=None,
        mapping_options={},
        cleanup_options={},
        dependencies=(),
    ):
        # Setting IN_FILE_2 to None makes AtomicCmd ignore this key
        aln = _bowtie2_template(
            ("bowtie2",),
            prefix,
            OUT_STDOUT=AtomicCmd.PIPE,
            CHECK_VERSION=BOWTIE2_VERSION,
        )

        aln.set_option("-x", prefix)

        if log_file is not None:
            aln.set_kwargs(OUT_STDERR=log_file)

        if input_file_1 and not input_file_2:
            aln.add_option("-U", input_file_1)
        elif input_file_1 and input_file_2:
            aln.add_option("-1", input_file_1)
            aln.add_option("-2", input_file_2)
        else:
            raise NodeError(
                "Input 1, OR both input 1 and input 2 must "
                "be specified for Bowtie2 node"
            )

        max_threads = _get_max_threads(reference, threads)
        aln.set_option("--threads", max_threads)

        cleanup = _new_cleanup_command(
            aln, output_file, reference, paired_end=input_file_1 and input_file_2
        )

        apply_options(aln, mapping_options)
        apply_options(cleanup, cleanup_options)

        algorithm = "PE" if input_file_2 else "SE"
        description = _get_node_description(
            name="Bowtie2",
            algorithm=algorithm,
            input_files_1=input_file_1,
            input_files_2=input_file_2,
            prefix=prefix,
            threads=threads,
        )

        CommandNode.__init__(
            self,
            command=ParallelCmds([aln.finalize(), cleanup.finalize()]),
            description=description,
            threads=threads,
            dependencies=dependencies,
        )


def _bowtie2_template(call, prefix, iotype="IN", **kwargs):
    for postfix in ("1.bt2", "2.bt2", "3.bt2", "4.bt2", "rev.1.bt2", "rev.2.bt2"):
        key = "%s_PREFIX_%s" % (iotype, postfix.upper())
        kwargs[key] = prefix + "." + postfix

    return AtomicCmdBuilder(call, **kwargs)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import argparse
import os
import sys

import paleomix.common.logging
import paleomix.resources
import paleomix.yaml


def build_parser():
    parser = argparse.ArgumentParser(prog="paleomix phylo_pipeline")

    parser.add_argument(
        "destination",
        default=".",
        nargs="?",
        help="Destination folder for example data.",
    )

    paleomix.common.logging.add_argument_group(parser)

    return parser


def main(argv):
    parser = build_parser()
    config = parser.parse_args(argv)
    paleomix.common.logging.initialize_console_logging()

    if paleomix.resources.copy_example("phylo_pipeline", config):
        return 1

    # Update interpreter to match the one currently in use;
    # this is required since we may be running from a virtual env
    filename = os.path.join(config.destination, "phylo_pipeline", "synthesize_reads.py")

    with open(filename) as handle:
        header, lines = handle.read().split("\n", 1)

    with open(filename, "w") as handle:
        handle.write("#!%s\n" % (os.path.abspath(sys.executable)))
        handle.write(lines)

    return 0



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import itertools

# BAM flags as defined in the BAM specification
BAM_SUPPLEMENTARY_ALIGNMENT = 0x800
BAM_PCR_DUPLICATE = 0x400
BAM_QUALITY_CONTROL_FAILED = 0x200
BAM_SECONDARY_ALIGNMENT = 0x100
BAM_IS_LAST_SEGMENT = 0x80
BAM_IS_FIRST_SEGMENT = 0x40
BAM_NEXT_IS_REVERSED = 0x20
BAM_READ_IS_REVERSED = 0x10
BAM_NEXT_IS_UNMAPPED = 0x8
BAM_READ_IS_UNMAPPED = 0x4
BAM_PROPER_SEGMENTS = 0x2
BAM_SEGMENTED = 0x1

# Default filters when processing reads
EXCLUDED_FLAGS = (
    BAM_SUPPLEMENTARY_ALIGNMENT
    | BAM_PCR_DUPLICATE
    | BAM_QUALITY_CONTROL_FAILED
    | BAM_SECONDARY_ALIGNMENT
    | BAM_READ_IS_UNMAPPED
)


class BAMRegionsIter:
    """Iterates over a BAM file, yield a separate iterator for each contig
    in the BAM or region in the list of regions if these are species, which in
    turn iterates over individual positions. This allows for the following
    pattern when parsing BAM files:

    for region in BAMRegionsIter(handle):
        # Setup per region
        for (position, records) in region:
            # Setup per position
            ...
            # Teardown per position
        # Teardown per region

    The list of regions given to the iterator is expected to be in BED-like
    records (see e.g. paleomix.common.bedtools), with these properties:
      - contig: Name of the contig in the BED file
      - start: 0-based offset for the start of the region
      - end: 1-based offset (i.e. past-the-end) of the region
      - name: The name of the region
    """

    def __init__(self, handle, regions=None, exclude_flags=EXCLUDED_FLAGS):
        """
          - handle: BAM file handle (c.f. module 'pysam')
          - regions: List of BED-like regions (see above)
        """
        self._handle = handle
        self._regions = regions
        self._excluded = exclude_flags

    def __iter__(self):
        if self._regions:
            for region in self._regions:
                records = self._handle.fetch(region.contig, region.start, region.end)
                records = self._filter(records)

                tid = self._handle.gettid(region.contig)
                yield _BAMRegion(tid, records, region.name, region.start, region.end)
        else:

            def _by_tid(record):
                """Group by reference ID."""
                return record.tid

            # Save a copy, as these are properties generated upon every access!
            names = self._handle.references
            lengths = self._handle.lengths
            records = self._filter(self._handle)
            records = itertools.groupby(records, key=_by_tid)

            for (tid, items) in records:
                if tid >= 0:
                    name = names[tid]
                    length = lengths[tid]
                else:
                    name = length = None

                yield _BAMRegion(tid, items, name, 0, length)

    def _filter(self, records):
        """Filters records by flags, if 'exclude_flags' is set."""
        if self._excluded:
            pred = lambda record: not record.flag & self._excluded
            return filter(pred, records)
        return records


class _BAMRegion:
    """Implements iteration over sites in a BAM file. It is assumed that the
    BAM file is sorted, and that the input records are from one contig.
    """

    def __init__(self, tid, records, name, start, end):
        self._records = records
        self.tid = tid
        self.name = name
        self.start = start
        self.end = end

    def __iter__(self):
        def _by_pos(record):
            """Group by position."""
            return record.pos

        for group in itertools.groupby(self._records, _by_pos):
            yield group



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy

from paleomix.node import Node
from paleomix.common.fileutils import move_file, reroot_path, describe_files
from paleomix.common.formats.msa import MSA
from paleomix.common.formats.phylip import interleaved_phy

from paleomix.common.utilities import safe_coerce_to_frozenset, safe_coerce_to_tuple


_VALID_KEYS = frozenset(["partitions", "filenames"])


class FastaToPartitionedInterleavedPhyNode(Node):
    def __init__(
        self,
        infiles,
        out_prefix,
        exclude_groups=(),
        reduce=False,
        dependencies=(),
        file_dependencies=(),
    ):
        """
        infiles = {names : {"partitions" : ..., "filenames" : [...]}}
        """
        if not (
            isinstance(infiles, dict)
            and all(isinstance(dd, dict) for dd in infiles.values())
        ):
            raise TypeError("'infiles' must be a dictionary of dictionaries")

        input_filenames = []
        for (name, subdd) in infiles.items():
            if set(subdd) - _VALID_KEYS:
                raise ValueError(
                    "Invalid keys found for %r: %s"
                    % (name, ", ".join(set(subdd) - _VALID_KEYS))
                )
            elif not isinstance(subdd["filenames"], list):
                raise ValueError("filenames must be a list of strings")
            input_filenames.extend(subdd["filenames"])
        # Optional file dependencies; used to depend on the list of sequcences
        input_filenames.extend(safe_coerce_to_tuple(file_dependencies))

        self._reduce = bool(reduce)
        self._infiles = copy.deepcopy(infiles)
        self._out_prefix = out_prefix
        self._excluded = safe_coerce_to_frozenset(exclude_groups)

        description = "creating%spartitioned phy from %s" % (
            "  reduced, " if reduce else " ",
            describe_files(input_filenames),
        )

        Node.__init__(
            self,
            description=description,
            input_files=input_filenames,
            output_files=[out_prefix + ".phy", out_prefix + ".partitions"],
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        merged_msas = []
        for (name, files_dd) in sorted(self._infiles.items()):
            partitions = files_dd["partitions"]
            msas = dict((key, []) for key in partitions)
            for filename in files_dd["filenames"]:
                msa = MSA.from_file(filename)
                if self._excluded:
                    msa = msa.exclude(self._excluded)

                for (key, msa_part) in msa.split(partitions).items():
                    msas[key].append(msa_part)

            msas.pop("X", None)
            for (key, msa_parts) in sorted(msas.items()):
                merged_msa = MSA.join(*msa_parts)
                if self._reduce:
                    merged_msa = merged_msa.reduce()

                if merged_msa is not None:
                    merged_msas.append(("%s_%s" % (name, key), merged_msa))

        out_fname_phy = reroot_path(temp, self._out_prefix + ".phy")
        with open(out_fname_phy, "w") as output_phy:
            final_msa = MSA.join(*(msa for (_, msa) in merged_msas))
            output_phy.write(interleaved_phy(final_msa))

        partition_end = 0
        out_fname_parts = reroot_path(temp, self._out_prefix + ".partitions")
        with open(out_fname_parts, "w") as output_part:
            for (name, msa) in merged_msas:
                length = msa.seqlen()
                output_part.write(
                    "DNA, %s = %i-%i\n"
                    % (name, partition_end + 1, partition_end + length)
                )
                partition_end += length

    def _teardown(self, _config, temp):
        move_file(
            reroot_path(temp, self._out_prefix + ".phy"), self._out_prefix + ".phy"
        )
        move_file(
            reroot_path(temp, self._out_prefix + ".partitions"),
            self._out_prefix + ".partitions",
        )



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from unittest.mock import call, Mock

import pytest

import paleomix.atomiccmd.pprint
from paleomix.atomiccmd.command import AtomicCmd, CmdError
from paleomix.atomiccmd.sets import ParallelCmds, SequentialCmds

_SET_CLASSES = (ParallelCmds, SequentialCmds)

###############################################################################
###############################################################################
# Properties with same expected behavior for both Parallel/SequentialCmds


@pytest.mark.parametrize("cls", _SET_CLASSES)
def test_atomicsets__properties(cls):
    cmd_mock_1 = AtomicCmd(
        ("true",),
        CHECK_A=id,
        EXEC_1="false",
        IN_1="/foo/bar/in_1.file",
        IN_2="/foo/bar/in_2.file",
        OUT_1="/bar/foo/out",
        TEMP_OUT_1="out.log",
        AUX_A="/aux/fA",
        AUX_B="/aux/fB",
    )
    cmd_mock_2 = AtomicCmd(
        ("false",),
        CHECK_A=list,
        EXEC_1="echo",
        EXEC_2="java",
        IN_1="/foo/bar/in.file",
        OUT_1="out.txt",
    )

    obj = cls([cmd_mock_1, cmd_mock_2])
    assert obj.executables == cmd_mock_1.executables | cmd_mock_2.executables
    assert obj.requirements == cmd_mock_1.requirements | cmd_mock_2.requirements
    assert obj.input_files == cmd_mock_1.input_files | cmd_mock_2.input_files
    assert obj.output_files == cmd_mock_1.output_files | cmd_mock_2.output_files
    assert (
        obj.auxiliary_files == cmd_mock_1.auxiliary_files | cmd_mock_2.auxiliary_files
    )
    assert obj.expected_temp_files == frozenset(["out", "out.txt"])
    assert (
        obj.optional_temp_files
        == cmd_mock_1.optional_temp_files | cmd_mock_2.optional_temp_files
    )


_NO_CLOBBERING_KWARGS = (
    ({"OUT_A": "/foo/out.txt"}, {"OUT_B": "/bar/out.txt"}),
    ({"OUT_A": "/foo/out.txt"}, {"TEMP_OUT_B": "out.txt"}),
    ({"OUT_A": "/foo/out.txt"}, {"OUT_STDOUT": "/bar/out.txt"}),
    ({"OUT_A": "/foo/out.txt"}, {"TEMP_OUT_STDOUT": "out.txt"}),
    ({"OUT_A": "/foo/out.txt"}, {"OUT_STDERR": "/bar/out.txt"}),
    ({"OUT_A": "/foo/out.txt"}, {"TEMP_OUT_STDERR": "out.txt"}),
)

# Ensure that commands in a set doesn't clobber eachothers OUT files
@pytest.mark.parametrize("cls", _SET_CLASSES)
@pytest.mark.parametrize("kwargs_1, kwargs_2", _NO_CLOBBERING_KWARGS)
def test_atomicsets__no_clobbering(cls, kwargs_1, kwargs_2):
    cmd_1 = AtomicCmd("true", **kwargs_1)
    cmd_2 = AtomicCmd("true", **kwargs_2)
    with pytest.raises(CmdError):
        cls([cmd_1, cmd_2])


###############################################################################
###############################################################################
# Functions with same expected behavior for both Parallel/SequentialCmds


@pytest.mark.parametrize("cls", _SET_CLASSES)
def test_atomicsets__commit(cls):
    mock = Mock()
    cmd_1 = AtomicCmd(["ls"])
    cmd_1.commit = mock.commit_1
    cmd_2 = AtomicCmd(["ls"])
    cmd_2.commit = mock.commit_2
    cmd_3 = AtomicCmd(["ls"])
    cmd_3.commit = mock.commit_3

    cls((cmd_1, cmd_2, cmd_3)).commit("xTMPx")

    assert mock.mock_calls == [
        call.commit_1("xTMPx"),
        call.commit_2("xTMPx"),
        call.commit_3("xTMPx"),
    ]


@pytest.mark.parametrize("cls", _SET_CLASSES)
def test_atomicsets__commit__remove_files_on_failure(tmp_path, cls):
    (tmp_path / "tmp").mkdir()
    out_path = tmp_path / "out"

    cmd_1 = AtomicCmd(["touch", "%(OUT_FILE)s"], OUT_FILE=str(out_path / "file1"))
    cmd_2 = AtomicCmd(["touch", "%(OUT_FILE)s"], OUT_FILE=str(out_path / "file2"))
    cmd_2.commit = Mock()
    cmd_2.commit.side_effect = OSError()

    cmdset = cls((cmd_1, cmd_2))
    cmdset.run(str(tmp_path / "tmp"))
    assert cmdset.join() == [0, 0]

    with pytest.raises(OSError):
        cmdset.commit(str(tmp_path / "tmp"))

    tmp_files = [it.name for it in (tmp_path / "tmp").iterdir()]
    assert "file1" not in tmp_files
    assert "file2" in tmp_files

    assert list((tmp_path / "out").iterdir()) == []


@pytest.mark.parametrize("cls", _SET_CLASSES)
def test_atomicsets__stdout(cls):
    cmds = cls([AtomicCmd("ls")])
    with pytest.raises(CmdError):
        cmds.stdout


@pytest.mark.parametrize("cls", _SET_CLASSES)
def test_atomicsets__terminate(cls):
    mock = Mock()
    cmd_1 = AtomicCmd(["ls"])
    cmd_1.terminate = mock.terminate_1
    cmd_2 = AtomicCmd(["ls"])
    cmd_2.terminate = mock.terminate_2
    cmd_3 = AtomicCmd(["ls"])
    cmd_3.terminate = mock.terminate_3

    cmds = cls((cmd_3, cmd_2, cmd_1))
    cmds.terminate()

    assert mock.mock_calls == [
        call.terminate_3(),
        call.terminate_2(),
        call.terminate_1(),
    ]


@pytest.mark.parametrize("cls", _SET_CLASSES)
def test_atomicsets__str__(cls):
    cmds = cls([AtomicCmd("ls")])
    assert paleomix.atomiccmd.pprint.pformat(cmds) == str(cmds)


@pytest.mark.parametrize("cls", _SET_CLASSES)
def test_atomicsets__duplicate_cmds(cls):
    cmd_1 = AtomicCmd("true")
    cmd_2 = AtomicCmd("false")
    with pytest.raises(ValueError):
        cls([cmd_1, cmd_2, cmd_1])


###############################################################################
###############################################################################
# Parallel commands


def test_parallel_commands__run():
    mock = Mock()
    cmd_1 = AtomicCmd(["ls"])
    cmd_1.run = mock.run_1
    cmd_2 = AtomicCmd(["ls"])
    cmd_2.run = mock.run_2
    cmd_3 = AtomicCmd(["ls"])
    cmd_3.run = mock.run_3

    cmds = ParallelCmds((cmd_1, cmd_2, cmd_3))
    cmds.run("xTMPx")

    assert mock.mock_calls == [
        call.run_1("xTMPx"),
        call.run_2("xTMPx"),
        call.run_3("xTMPx"),
    ]


@pytest.mark.parametrize("value", (True, False))
def test_parallel_commands__ready_single(value):
    cmd = AtomicCmd(["ls"])
    cmd.ready = Mock()
    cmd.ready.return_value = value
    cmds = ParallelCmds([cmd])
    assert cmds.ready() == value

    assert cmd.ready.mock_calls


_READY_TWO_VALUES = (
    (True, True, True),
    (False, True, False),
    (True, False, False),
    (False, False, False),
)


@pytest.mark.parametrize("first, second, result", _READY_TWO_VALUES)
def test_parallel_commands__ready_two(first, second, result):
    cmd_1 = AtomicCmd(["ls"])
    cmd_1.ready = Mock()
    cmd_1.ready.return_value = first

    cmd_2 = AtomicCmd(["ls"])
    cmd_2.ready = Mock()
    cmd_2.ready.return_value = second
    cmds = ParallelCmds([cmd_1, cmd_2])
    assert cmds.ready() == result

    assert cmd_1.ready.mock_calls
    assert bool(first) == bool(cmd_2.ready.call_count)


def test_parallel_commands__join_before_run():
    mock = Mock()
    cmd_1 = AtomicCmd(["ls"])
    cmd_1.join = mock.join_1
    cmd_2 = AtomicCmd(["ls"])
    cmd_2.join = mock.join_2
    cmd_3 = AtomicCmd(["ls"])
    cmd_3.join = mock.join_3

    cmds = ParallelCmds((cmd_3, cmd_2, cmd_1))
    assert cmds.join() == [None, None, None]

    assert mock.mock_calls == []


def test_parallel_commands__join_after_run(tmp_path):
    cmds = ParallelCmds([AtomicCmd("true") for _ in range(3)])
    cmds.run(str(tmp_path))
    assert cmds.join() == [0, 0, 0]


def _setup_mocks_for_failure(*do_mocks):
    results = []
    for do_mock in do_mocks:
        if do_mock:
            mock = AtomicCmd(("sleep", 10))
        else:
            mock = AtomicCmd("false")
        results.append(mock)
    return results


def test_parallel_commands__join_failure_1(tmp_path):
    mocks = _setup_mocks_for_failure(False, True, True)
    cmds = ParallelCmds(mocks)
    cmds.run(str(tmp_path))
    assert cmds.join() == [1, "SIGTERM", "SIGTERM"]


def test_parallel_commands__join_failure_2(tmp_path):
    mocks = _setup_mocks_for_failure(True, False, True)
    cmds = ParallelCmds(mocks)
    cmds.run(str(tmp_path))
    assert cmds.join() == ["SIGTERM", 1, "SIGTERM"]


def test_parallel_commands__join_failure_3(tmp_path):
    mocks = _setup_mocks_for_failure(True, True, False)
    cmds = ParallelCmds(mocks)
    cmds.run(str(tmp_path))
    assert cmds.join() == ["SIGTERM", "SIGTERM", 1]


def test_parallel_commands__reject_sequential():
    command = AtomicCmd(["ls"])
    seqcmd = SequentialCmds([command])
    with pytest.raises(CmdError):
        ParallelCmds([seqcmd])


def test_parallel_commands__accept_parallel():
    command = AtomicCmd(["ls"])
    parcmd = ParallelCmds([command])
    ParallelCmds([parcmd])


def test_parallel_commands__reject_noncommand():
    with pytest.raises(CmdError):
        ParallelCmds([object()])


def test_parallel_commands__reject_empty_commandset():
    with pytest.raises(CmdError):
        ParallelCmds([])


###############################################################################
###############################################################################
# Sequential commands


def test_sequential_commands__atomiccmds():
    mock = Mock()
    cmd_1 = AtomicCmd(["ls"])
    cmd_1.run = mock.run_1
    cmd_1.join = mock.join_1
    cmd_1.join.return_value = [0]
    cmd_2 = AtomicCmd(["ls"])
    cmd_2.run = mock.run_2
    cmd_2.join = mock.join_2
    cmd_2.join.return_value = [0]
    cmd_3 = AtomicCmd(["ls"])
    cmd_3.run = mock.run_3
    cmd_3.join = mock.join_3
    cmd_3.join.return_value = [0]

    cmds = SequentialCmds((cmd_1, cmd_2, cmd_3))
    assert not cmds.ready()
    cmds.run("xTMPx")
    assert cmds.ready()
    assert cmds.join() == [0, 0, 0]

    assert mock.mock_calls == [
        call.run_1("xTMPx"),
        call.join_1(),
        call.run_2("xTMPx"),
        call.join_2(),
        call.run_3("xTMPx"),
        call.join_3(),
        call.join_1(),
        call.join_2(),
        call.join_3(),
    ]


def test_sequential_commands__abort_on_error_1(tmp_path):
    cmd_1 = AtomicCmd("false")
    cmd_2 = AtomicCmd(("sleep", 10))
    cmd_3 = AtomicCmd(("sleep", 10))
    cmds = SequentialCmds([cmd_1, cmd_2, cmd_3])
    cmds.run(str(tmp_path))
    assert cmds.join() == [1, None, None]


def test_sequential_commands__abort_on_error_2(tmp_path):
    cmd_1 = AtomicCmd("true")
    cmd_2 = AtomicCmd("false")
    cmd_3 = AtomicCmd(("sleep", 10))
    cmds = SequentialCmds([cmd_1, cmd_2, cmd_3])
    cmds.run(str(tmp_path))
    assert cmds.join() == [0, 1, None]


def test_sequential_commands__abort_on_error_3(tmp_path):
    cmd_1 = AtomicCmd("true")
    cmd_2 = AtomicCmd("true")
    cmd_3 = AtomicCmd("false")
    cmds = SequentialCmds([cmd_1, cmd_2, cmd_3])
    cmds.run(str(tmp_path))
    assert cmds.join() == [0, 0, 1]


def test_sequential_commands__accept_parallel():
    command = AtomicCmd(["ls"])
    parcmd = ParallelCmds([command])
    SequentialCmds([parcmd])


def test_sequential_commands__accept_sequential():
    command = AtomicCmd(["ls"])
    seqcmd = SequentialCmds([command])
    SequentialCmds([seqcmd])


def test_sequential_commands__reject_noncommand():
    with pytest.raises(CmdError):
        SequentialCmds([object()])


def test_sequential_commands__reject_empty_commandset():
    with pytest.raises(CmdError):
        SequentialCmds([])



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import glob
import logging
import os
import sys


_TEMPLATE_TOP = """# -*- mode: Yaml; -*-
# Default options.
# Can also be specific for a set of samples, libraries, and lanes,
# by including the "Options" hierarchy at the same level as those
# samples, libraries, or lanes below.
Options:
  # Sequencing platform, see SAM/BAM reference for valid values
  Platform: Illumina
  # Quality offset for Phred scores, either 33 (Sanger/Illumina 1.8+)
  # or 64 (Illumina 1.3+ / 1.5+). For Bowtie2 it is also possible to
  # specify 'Solexa', to handle reads on the Solexa scale. This is
  # used during adapter-trimming and sequence alignment
  QualityOffset: 33

  # Settings for trimming of reads, see AdapterRemoval man-page
  AdapterRemoval:
     # Set and uncomment to override defaults adapter sequences
#     --adapter1: AGATCGGAAGAGCACACGTCTGAACTCCAGTCACNNNNNNATCTCGTATGCCGTCTTCTGCTTG
#     --adapter2: AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT
     # Some BAM pipeline defaults differ from AR defaults;
     # To override, change these value(s):
     --mm: 3
     --minlength: 25
     # Extra features enabled by default; change 'yes' to 'no' to disable
     --collapse: yes
     --trimns: yes
     --trimqualities: yes
     # Increase the maximum Phred allowed for input FASTQs, as well as for merged bases
     # when using --collapse (default = 41). This is needed for some modern FASTQs.
#     --qualitymax: 42
"""

_TEMPLATE_BAM_OPTIONS = """  # Settings for aligners supported by the pipeline
  Aligners:
    # Choice of aligner software to use, either "BWA" or "Bowtie2"
    Program: BWA

    # Settings for mappings performed using BWA
    BWA:
      # One of "backtrack", "bwasw", or "mem"; see the BWA documentation
      # for a description of each algorithm (defaults to 'backtrack')
      Algorithm: backtrack
      # Filter aligned reads with a mapping quality (Phred) below this value
      MinQuality: 0
      # Filter reads that did not map to the reference sequence
      FilterUnmappedReads: yes
      # May be disabled ("no") for aDNA alignments with the 'aln' algorithm.
      # Post-mortem damage localizes to the seed region, which BWA expects to
      # have few errors (sets "-l"). See http://pmid.us/22574660
      UseSeed: yes
      # Additional command-line options may be specified below. For 'backtrack' these
      # are applied to the "bwa aln". See Bowtie2 for more examples.
#      -n: 0.04

    # Settings for mappings performed using Bowtie2
    Bowtie2:
      # Filter aligned reads with a mapping quality (Phred) below this value
      MinQuality: 0
      # Filter reads that did not map to the reference sequence
      FilterUnmappedReads: yes
      # Examples of how to add additional command-line options
#      --trim5: 5
#      --trim3: 5
      # Note that the colon is required, even if no value is specified
      --very-sensitive:
      # Example of how to specify multiple values for an option
#      --rg:
#        - CN:SequencingCenterNameHere
#        - DS:DescriptionOfReadGroup

  # Command-line options for mapDamage; use long-form options(--length not -l):
  mapDamage:
    # By default, the pipeline will downsample the input to 100k hits
    # when running mapDamage; remove to use all hits
    --downsample: 100000

  # Set to 'yes' exclude a type of trimmed reads from alignment / analysis;
  # possible read-types reflect the output of AdapterRemoval
  ExcludeReads:
    # Exclude single-end reads (yes / no)?
    Single: no
    # Exclude non-collapsed paired-end reads (yes / no)?
    Paired: no
    # Exclude paired-end reads for which the mate was discarded (yes / no)?
    Singleton: no
    # Exclude overlapping paired-ended reads collapsed into a single sequence
    # by AdapterRemoval (yes / no)?
    Collapsed: no
    # Like 'Collapsed', but only for collapsed reads truncated due to the
    # presence of ambiguous or low quality bases at read termini (yes / no).
    CollapsedTruncated: no

  # Optional steps to perform during processing.
  Features:
    # If set to 'filter', PCR duplicates are removed from the output files; if set to
    # 'mark', PCR duplicates are flagged with bit 0x400, and not removed from the
    # output files; if set to 'no', the reads are assumed to not have been amplified.
    PCRDuplicates: filter
    # Set to 'no' to disable mapDamage; set to 'plots' to build basic mapDamage plots;
    # set to 'model' to build plots and post-mortem damage models; and set to 'rescale'
    # to build plots, models, and BAMs with rescaled quality scores. All analyses are
    # carried out per library.
    mapDamage: plot
    # Generate coverage information for the final BAM and for each 'RegionsOfInterest'
    # specified in 'Prefixes' (yes / no).
    Coverage: yes
    # Generate histograms of number of sites with a given read-depth, from 0 to 200,
    # for each BAM and for each 'RegionsOfInterest' specified in 'Prefixes' (yes / no).
    Depths: yes
    # Generate summary table for each target (yes / no)
    Summary: yes
"""

_TEMPLATE_PREFIXES = """
# Map of prefixes by name, each having a Path, which specifies the location of the
# BWA/Bowtie2 index, and optional regions for which additional statistics are produced.
Prefixes:
  # Replace 'NAME_OF_PREFIX' with name of the prefix; this name is used in summary
  # statistics and as part of output filenames.
  NAME_OF_PREFIX:
    # Replace 'PATH_TO_PREFIX' with the path to .fasta file containing the references
    # against which reads are to be mapped. Using the same name as filename is strongly
    # recommended (e.g. /path/to/Human_g1k_v37.fasta should be named 'Human_g1k_v37').
    Path: PATH_TO_PREFIX.fasta

    # (Optional) Uncomment and replace 'PATH_TO_BEDFILE' with the path to a .bed file
    # listing extra regions for which coverage / depth statistics should be calculated;
    # if no names are specified for the BED records, results are named after the
    # chromosome / contig. Replace 'NAME' with the desired name for these regions.
#    RegionsOfInterest:
#      NAME: PATH_TO_BEDFILE
"""

_TEMPLATE_SAMPLES = """
# Mapping targets are specified using the following structure. Replace 'NAME_OF_TARGET'
# with the desired prefix for filenames.
NAME_OF_TARGET:
  # Replace 'NAME_OF_SAMPLE' with the name of this sample.
  NAME_OF_SAMPLE:
    # Replace 'NAME_OF_LIBRARY' with the name of this sample.
    NAME_OF_LIBRARY:
       # Replace 'NAME_OF_LANE' with the lane name (e.g. the barcode) and replace
       # 'PATH_WITH_WILDCARDS' with the path to the FASTQ files to be trimmed and mapped
       # for this lane (may include wildcards).
      NAME_OF_LANE: PATH_WITH_WILDCARDS
"""

_FILENAME = "SampleSheet.csv"


def build_makefile(add_full_options=True, add_prefix_tmpl=True, add_sample_tmpl=True):
    template_parts = [_TEMPLATE_TOP]

    if add_full_options:
        template_parts.append(_TEMPLATE_BAM_OPTIONS)

    if add_prefix_tmpl:
        template_parts.append(_TEMPLATE_PREFIXES)

    if add_sample_tmpl:
        template_parts.append(_TEMPLATE_SAMPLES)

    return "\n".join(template_parts)


def strip_comments(text):
    lines = text.split("\n")

    # Always include minimal header
    minimal_template = lines[:3]
    for line in lines[3:]:
        if not line.lstrip().startswith("#"):
            line = line.split("#", 1)[0].rstrip()

            # Avoid too many empty lines
            if line.strip() or minimal_template[-1].strip():
                minimal_template.append(line)

    return "\n".join(minimal_template)


def read_alignment_records(filename):
    log = logging.getLogger(__name__)

    results = []
    with open(filename) as records:
        line = records.readline()
        if not line:
            log.error("Empty SampleSheet.csv file: %r", filename)
            return None

        header = line.strip().split(",")
        missing = set(("SampleID", "Index", "Lane", "FCID")) - set(header)
        if missing:
            log.error(
                "Required columns missing from SampleSheet file %r: %s",
                filename,
                ", ".join(map(repr, missing)),
            )
            return None

        for idx, line in enumerate(records, start=2):
            line = line.strip()
            if not line:
                continue

            fields = line.split(",")
            if len(fields) != len(header):
                log.error(
                    "Line %i in SampleSheet file %r does not contain the expected "
                    "number of columns; expected %i, but found %i.",
                    idx,
                    filename,
                    len(header),
                    len(fields),
                )
                return None

            results.append(dict(zip(header, fields)))

    return results


def select_path(path):
    has_r1 = bool(glob.glob(path.format(Pair=1)))
    has_r2 = bool(glob.glob(path.format(Pair=2)))

    if has_r1 and not has_r2:
        # Single-ended reads
        return path.format(Pair=1)
    return path


def read_sample_sheets(filenames):
    records = {}
    for root in filenames:
        if os.path.isdir(root):
            filename = os.path.join(root, _FILENAME)
        else:
            root, filename = os.path.split(root)[0], root

        if not os.path.exists(filename):
            log = logging.getLogger(__name__)
            log.error("Could not find SampleSheet file: %r", filename)
            return None

        sample_sheet = read_alignment_records(filename)
        if sample_sheet is None:
            return None

        for record in sample_sheet:
            record["Lane"] = int(record["Lane"])
            path = "%(SampleID)s_%(Index)s_L%(Lane)03i_R{Pair}_*.fastq.gz" % record
            record["Path"] = select_path(os.path.join(root, path))
            key = "%(FCID)s_%(Lane)s" % record

            libraries = records.setdefault(record["SampleID"], {})
            barcodes = libraries.setdefault(record["Index"], {})
            barcodes.setdefault(key, []).append(path)

    # Clean up names; generate unique names for duplicate lanes
    for libraries in records.values():
        for barcodes in libraries.values():
            for key, paths in list(barcodes.items()):
                if len(paths) == 1:
                    barcodes[key] = paths[0]
                    continue

                counter = 1
                for path in paths:
                    new_key = "%s_%i" % (key, counter)

                    while new_key in barcodes:
                        counter += 1
                        new_key = "%s_%i" % (key, counter)

                    barcodes[new_key] = path

                barcodes.pop(key)

    return records


def print_samples(records):
    print()
    for (sample, libraries) in sorted(records.items()):
        print("%s:" % sample)
        print("  %s:" % sample)
        for (library, barcodes) in sorted(libraries.items()):
            print("    %s:" % library)
            for key, path in sorted(barcodes.items()):
                print("      %s: %s" % (key, path))
            print()
        print()


def main(args, pipeline="bam"):
    if pipeline not in ("bam", "trim"):
        raise ValueError(pipeline)

    log = logging.getLogger(__name__)
    records = read_sample_sheets(args.samplesheets)
    if records is None:
        return 1

    template = build_makefile(
        add_full_options=(pipeline == "bam"),
        add_prefix_tmpl=(pipeline == "bam"),
        add_sample_tmpl=not records,
    )
    if args.minimal:
        template = strip_comments(template)

    print(template.rstrip())

    print_samples(records)

    if args.samplesheets:
        log.info("Automatically generated makefile printed.")
        log.info("Please check for correctness before running pipeline.")
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import codecs
import os
import sys

from setuptools import setup, find_packages

if sys.version_info < (3, 5):
    sys.stderr.write("FATAL ERROR: ")
    sys.stderr.write("PALEOMIX requires at least Python 3.5, but setup.py ")
    sys.stderr.write("was run using Python %s.%s!\n" % sys.version_info[:2])
    sys.exit(1)


def _get_version():
    """Retrieve version from current install directory."""
    env = {}
    with open(os.path.join("paleomix", "__init__.py")) as handle:
        exec(handle.read(), env)

    return env["__version__"]


def _get_readme():
    """Retrieves contents of README.rst, forcing UTF-8 encoding."""
    with codecs.open("README.rst", encoding="utf-8") as handle:
        return handle.read()


setup(
    name="paleomix",
    version=_get_version(),
    description="Bioinformatics pipelines for HTS data",
    long_description=_get_readme(),
    url="https://github.com/MikkelSchubert/paleomix",
    author="Mikkel Schubert",
    author_email="MikkelSch@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
    ],
    keywords="pipeline bioinformatics hts phylogeny bam",
    packages=find_packages(exclude=["misc", "tests"]),
    install_requires=[
        "coloredlogs>=10.0",
        "configargparse>=0.13.0",
        "pysam>=0.10.0",
        "ruamel.yaml>=0.16.0",
        "setproctitle>=1.1.0",
    ],
    entry_points={"console_scripts": ["paleomix=paleomix.main:entry_point"]},
    zip_safe=False,
    include_package_data=True,
)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import signal

import pytest

from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import ParallelCmds, SequentialCmds
from paleomix.atomiccmd.pprint import pformat, _pformat_list


###############################################################################
###############################################################################
# pformat


def test_pformat__simple():
    cmd = AtomicCmd(("touch", "something"))
    assert pformat(cmd) == (
        "Command = touch something\n"
        "STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'"
    ) % (id(cmd), id(cmd))


def test_pformat__simple__running(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)
    assert pformat(cmd) == (
        "Command = sleep 10\n"
        "Status  = Running\n"
        "STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
        "STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
        "CWD     = '{cwd}'"
    ).format(id=id(cmd), cwd=os.getcwd(), temp_dir=tmp_path)
    cmd.terminate()
    cmd.join()


def test_pformat__simple__running__set_cwd(tmp_path):
    cmd = AtomicCmd(("sleep", "10"), set_cwd=True)
    cmd.run(tmp_path)
    assert pformat(cmd) == (
        "Command = sleep 10\n"
        "Status  = Running\n"
        "STDOUT* = 'pipe_sleep_{id}.stdout'\n"
        "STDERR* = 'pipe_sleep_{id}.stderr'\n"
        "CWD     = '{temp_dir}'"
    ).format(id=id(cmd), temp_dir=tmp_path)
    cmd.terminate()
    cmd.join()


def test_pformat__simple__done(tmp_path):
    cmd = AtomicCmd("true")
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    assert pformat(cmd) == (
        "Command = true\n"
        "Status  = Exited with return-code 0\n"
        "STDOUT* = '{temp_dir}/pipe_true_{id}.stdout'\n"
        "STDERR* = '{temp_dir}/pipe_true_{id}.stderr'\n"
        "CWD     = '{cwd}'"
    ).format(id=id(cmd), cwd=os.getcwd(), temp_dir=tmp_path)


def test_pformat__simple__done__before_join(tmp_path):
    cmd = AtomicCmd("true")
    cmd.run(tmp_path)
    cmd._proc.wait()
    assert pformat(cmd) == (
        "Command = true\n"
        "Status  = Exited with return-code 0\n"
        "STDOUT* = '{temp_dir}/pipe_true_{id}.stdout'\n"
        "STDERR* = '{temp_dir}/pipe_true_{id}.stderr'\n"
        "CWD     = '{cwd}'"
    ).format(id=id(cmd), cwd=os.getcwd(), temp_dir=tmp_path)
    assert cmd.join() == [0]


def test_pformat__simple__done__set_cwd(tmp_path):
    cmd = AtomicCmd("true", set_cwd=True)
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    assert pformat(cmd) == (
        "Command = true\n"
        "Status  = Exited with return-code 0\n"
        "STDOUT* = 'pipe_true_{id}.stdout'\n"
        "STDERR* = 'pipe_true_{id}.stderr'\n"
        "CWD     = '{temp_dir}'"
    ).format(id=id(cmd), temp_dir=tmp_path)


def test_pformat__simple__terminated_by_pipeline(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)
    cmd.terminate()
    assert cmd.join() == ["SIGTERM"]
    assert pformat(cmd) == (
        "Command = sleep 10\n"
        "Status  = Automatically terminated by PALEOMIX\n"
        "STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
        "STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
        "CWD     = '{cwd}'"
    ).format(id=id(cmd), temp_dir=tmp_path, cwd=os.getcwd())


def test_pformat__simple__killed_by_signal(tmp_path):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(tmp_path)
    os.killpg(cmd._proc.pid, signal.SIGTERM)
    assert cmd.join() == ["SIGTERM"]
    assert pformat(cmd) == (
        "Command = sleep 10\n"
        "Status  = Terminated with signal SIGTERM\n"
        "STDOUT* = '{temp_dir}/pipe_sleep_{id}.stdout'\n"
        "STDERR* = '{temp_dir}/pipe_sleep_{id}.stderr'\n"
        "CWD     = '{cwd}'"
    ).format(id=id(cmd), temp_dir=tmp_path, cwd=os.getcwd())


def test_pformat__simple__temp_root_in_arguments(tmp_path):
    cmd = AtomicCmd(("echo", "${TEMP_DIR}"))
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    assert pformat(cmd) == (
        "Command = echo '{temp_dir}'\n"
        "Status  = Exited with return-code 0\n"
        "STDOUT* = '{temp_dir}/pipe_echo_{id}.stdout'\n"
        "STDERR* = '{temp_dir}/pipe_echo_{id}.stderr'\n"
        "CWD     = '{cwd}'"
    ).format(id=id(cmd), temp_dir=tmp_path, cwd=os.getcwd())


###############################################################################
###############################################################################
# INFILE


def test_pformat__atomiccmd__simple_with_infile():
    cmd = AtomicCmd(("cat", "%(IN_SOMETHING)s"), IN_SOMETHING="/etc/fstab")
    assert pformat(cmd) == (
        "Command = cat /etc/fstab\n"
        "STDOUT* = '${TEMP_DIR}/pipe_cat_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_cat_%i.stderr'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_infile__set_cwd():
    cmd = AtomicCmd(
        ("cat", "%(IN_SOMETHING)s"), IN_SOMETHING="/etc/fstab", set_cwd=True
    )
    assert pformat(cmd) == (
        "Command = cat /etc/fstab\n"
        "STDOUT* = 'pipe_cat_%i.stdout'\n"
        "STDERR* = 'pipe_cat_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_temp_infile():
    cmd = AtomicCmd(("cat", "%(TEMP_IN_FILE)s"), TEMP_IN_FILE="infile.txt")

    assert pformat(cmd) == (
        "Command = cat '${TEMP_DIR}/infile.txt'\n"
        "STDOUT* = '${TEMP_DIR}/pipe_cat_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_cat_%i.stderr'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_temp_infile__set_cwd():
    cmd = AtomicCmd(
        ("zcat", "%(TEMP_IN_FILE)s"), TEMP_IN_FILE="infile.gz", set_cwd=True
    )

    assert pformat(cmd) == (
        "Command = zcat infile.gz\n"
        "STDOUT* = 'pipe_zcat_%i.stdout'\n"
        "STDERR* = 'pipe_zcat_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd), id(cmd))


###############################################################################
###############################################################################
# OUTFILE


def test_pformat__atomiccmd__simple_with_outfile():
    cmd = AtomicCmd(("touch", "%(OUT_RC)s"), OUT_RC="/etc/bashrc")
    assert pformat(cmd) == (
        "Command = touch '${TEMP_DIR}/bashrc'\n"
        "STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_outfile__set_cwd():
    cmd = AtomicCmd(("touch", "%(OUT_RC)s"), OUT_RC="/etc/bashrc", set_cwd=True)

    assert pformat(cmd) == (
        "Command = touch bashrc\n"
        "STDOUT* = 'pipe_touch_%i.stdout'\n"
        "STDERR* = 'pipe_touch_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_temp_outfile():
    cmd = AtomicCmd(("touch", "%(TEMP_OUT_RC)s"), TEMP_OUT_RC="bashrc")

    assert pformat(cmd) == (
        "Command = touch '${TEMP_DIR}/bashrc'\n"
        "STDOUT* = '${TEMP_DIR}/pipe_touch_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_touch_%i.stderr'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_temp_outfile__set_cwd():
    cmd = AtomicCmd(("touch", "%(TEMP_OUT_RC)s"), TEMP_OUT_RC="bashrc", set_cwd=True)

    assert pformat(cmd) == (
        "Command = touch bashrc\n"
        "STDOUT* = 'pipe_touch_%i.stdout'\n"
        "STDERR* = 'pipe_touch_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd), id(cmd))


###############################################################################
###############################################################################
# STDIN


def test_pformat__atomiccmd__simple_with_stdin():
    cmd = AtomicCmd("gzip", IN_STDIN="/etc/fstab")
    assert pformat(cmd) == (
        "Command = gzip\n"
        "STDIN   = '/etc/fstab'\n"
        "STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_stdin__set_cwd():
    cmd = AtomicCmd("gzip", IN_STDIN="/etc/fstab", set_cwd=True)
    assert pformat(cmd) == (
        "Command = gzip\n"
        "STDIN   = '/etc/fstab'\n"
        "STDOUT* = 'pipe_gzip_%i.stdout'\n"
        "STDERR* = 'pipe_gzip_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_temp_stdin():
    cmd = AtomicCmd("gzip", TEMP_IN_STDIN="stabstabstab")
    assert pformat(cmd) == (
        "Command = gzip\n"
        "STDIN*  = '${TEMP_DIR}/stabstabstab'\n"
        "STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_temp_stdin__set_cwd():
    cmd = AtomicCmd("gzip", TEMP_IN_STDIN="stabstabstab", set_cwd=True)
    assert pformat(cmd) == (
        "Command = gzip\n"
        "STDIN*  = 'stabstabstab'\n"
        "STDOUT* = 'pipe_gzip_%i.stdout'\n"
        "STDERR* = 'pipe_gzip_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd), id(cmd))


def test_pformat__atomiccmd__simple_with_stdin__cmd():
    cmd_1 = AtomicCmd("gzip", OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
    assert pformat(cmd_2) == (
        "Command = gzip\n"
        "STDIN   = <PIPE>\n"
        "STDOUT* = '${TEMP_DIR}/pipe_gzip_%i.stdout'\n"
        "STDERR* = '${TEMP_DIR}/pipe_gzip_%i.stderr'"
    ) % (id(cmd_2), id(cmd_2))


###############################################################################
###############################################################################
# STDOUT


def test_pformat__atomiccmd__simple_with_stdout():
    cmd = AtomicCmd(("echo", "Water. Water."), OUT_STDOUT="/dev/ls")
    assert pformat(cmd) == (
        "Command = echo 'Water. Water.'\n"
        "STDOUT  = '${TEMP_DIR}/ls'\n"
        "STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'"
    ) % (id(cmd),)


def test_pformat__atomiccmd__simple_with_stdout__set_cwd():
    cmd = AtomicCmd(("echo", "*pant*. *pant*."), OUT_STDOUT="/dev/barf", set_cwd=True)
    assert pformat(cmd) == (
        "Command = echo '*pant*. *pant*.'\n"
        "STDOUT  = 'barf'\n"
        "STDERR* = 'pipe_echo_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd),)


def test_pformat__atomiccmd__simple_with_temp_stdout():
    cmd = AtomicCmd(("echo", "Oil. Oil."), TEMP_OUT_STDOUT="dm")
    assert pformat(cmd) == (
        "Command = echo 'Oil. Oil.'\n"
        "STDOUT* = '${TEMP_DIR}/dm'\n"
        "STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'"
    ) % (id(cmd),)


def test_pformat__atomiccmd__simple_with_temp_stdout__set_cwd():
    cmd = AtomicCmd(
        ("echo", "Room service. Room service."), TEMP_OUT_STDOUT="pv", set_cwd=True
    )
    assert pformat(cmd) == (
        "Command = echo 'Room service. Room service.'\n"
        "STDOUT* = 'pv'\n"
        "STDERR* = 'pipe_echo_%i.stderr'\n"
        "CWD     = '${TEMP_DIR}'"
    ) % (id(cmd),)


def test_pformat__atomiccmd__simple_with_stdout_pipe():
    cmd = AtomicCmd(("echo", "!"), OUT_STDOUT=AtomicCmd.PIPE)
    assert pformat(cmd) == (
        "Command = echo '!'\n"
        "STDOUT  = <PIPE>\n"
        "STDERR* = '${TEMP_DIR}/pipe_echo_%i.stderr'"
    ) % (id(cmd),)


###############################################################################
###############################################################################
# ParallelCmds


@pytest.mark.parametrize(
    "cls, description",
    ((ParallelCmds, "Parallel processes"), (SequentialCmds, "Sequential processes")),
)
def test_pformat__sets__simple(cls, description):
    template = (
        "{description}:\n"
        "  Process 1:\n"
        "    Command = echo foo\n"
        "    STDOUT  = Piped to process 2\n"
        "    STDERR* = '${{TEMP_DIR}}/pipe_echo_{cmd_1_id}.stderr'\n"
        "\n"
        "  Process 2:\n"
        "    Command = gzip\n"
        "    STDIN   = Piped from process 1\n"
        "    STDOUT* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stdout'\n"
        "    STDERR* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2_id}.stderr'"
    )

    cmd_1 = AtomicCmd(("echo", "foo"), OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
    cmd = cls((cmd_1, cmd_2))
    assert pformat(cmd) == template.format(
        description=description, cmd_1_id=id(cmd_1), cmd_2_id=id(cmd_2)
    )


def test_pformat__sets__nested():
    cmd_1 = AtomicCmd(("echo", "foo"), OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2 = AtomicCmd("gzip", IN_STDIN=cmd_1)
    cmd_3 = AtomicCmd("sha1sum")
    set_1 = ParallelCmds((cmd_1, cmd_2))
    set_2 = SequentialCmds((set_1, cmd_3))
    assert pformat(set_2) == (
        "Sequential processes:\n"
        "  Parallel processes:\n"
        "    Process 1:\n"
        "      Command = echo foo\n"
        "      STDOUT  = Piped to process 2\n"
        "      STDERR* = '${{TEMP_DIR}}/pipe_echo_{cmd_1}.stderr'\n"
        "\n"
        "    Process 2:\n"
        "      Command = gzip\n"
        "      STDIN   = Piped from process 1\n"
        "      STDOUT* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2}.stdout'\n"
        "      STDERR* = '${{TEMP_DIR}}/pipe_gzip_{cmd_2}.stderr'\n"
        "\n"
        "  Process 3:\n"
        "    Command = sha1sum\n"
        "    STDOUT* = '${{TEMP_DIR}}/pipe_sha1sum_{cmd_3}.stdout'\n"
        "    STDERR* = '${{TEMP_DIR}}/pipe_sha1sum_{cmd_3}.stderr'"
    ).format(cmd_1=id(cmd_1), cmd_2=id(cmd_2), cmd_3=id(cmd_3))


###############################################################################
###############################################################################
# Bad input


@pytest.mark.parametrize("value", (1, {}, ""))
def test_pformat__bad_input(value):
    with pytest.raises(TypeError):
        pformat(value)


###############################################################################
###############################################################################
# _pformat_list


def test_pformat_list__empty():
    assert _pformat_list([]) == ""


def test_pformat_list__single():
    assert _pformat_list([3]) == "3"


def test_pformat_list__multiple():
    assert _pformat_list([3, 2, 1]) == "3 2 1"


def test_pformat_list__wrapped():
    assert _pformat_list([3, 2, 1], width=1) == "3 \\\n    2 \\\n    1"
    assert _pformat_list([3, 2, 1], width=2) == "3 \\\n    2 \\\n    1"
    assert _pformat_list([3, 2, 1], width=3) == "3 \\\n    2 \\\n    1"
    assert _pformat_list([3, 2, 1], width=4) == "3 2 \\\n    1"
    assert _pformat_list([3, 2, 1], width=5) == "3 2 \\\n    1"
    assert _pformat_list([3, 2, 1], width=6) == "3 2 1"
    assert _pformat_list([3, 2, 1], width=7) == "3 2 1"


def test_pformat_list__escaped():
    assert _pformat_list(["a", "b c"], width=100) == "a 'b c'"
    assert _pformat_list(["a", "$c"], width=100) == "a '$c'"
    assert _pformat_list(["!a", "c"], width=100) == "'!a' c"
    assert _pformat_list(["a", "'c"], width=100) == """a ''"'"'c'"""
